import dataclasses
import importlib
import inspect
import logging
import pkgutil
import re
import sys
import typing
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Optional, Union

import apischema
import blark
import pytest

from .. import (access_security, asyn, autosave, cache, common, ioc_finder,
                motor, shell, snl)
from ..common import FullLoadContext, LoadContext
from ..plugins import twincat_pytmc

MODULE_PATH = Path(__file__).parent

logger = logging.getLogger(__name__)

SKIP_CLASSES = (
    cache.Cached,
    ioc_finder.IocScriptStaticInfoList,
    ioc_finder.IocScriptStaticList,
    twincat_pytmc.PytmcMetadata,  # should not serialize
)

SKIP_DESERIALIZATION = {
    # These take too long to round-trip and verify somehow:
    shell.LoadedIoc,
    shell.ShellState,
    autosave.AutosaveState,
    asyn.AsynState,
    motor.MotorState,
    access_security.AccessSecurityState,
}


def find_whatrecord_submodules() -> Dict[str, ModuleType]:
    """Find all whatrecord submodules, as a dictionary of name to module."""
    modules = {}
    package_root = str(MODULE_PATH.parent)
    for item in pkgutil.walk_packages(path=[package_root], prefix="whatrecord."):
        if item.name.endswith("__main__"):
            continue
        try:
            modules[item.name] = sys.modules[item.name]
        except KeyError:
            # Submodules may not yet be imported; do that here.
            try:
                modules[item.name] = importlib.import_module(
                    item.name, package="whatrecord"
                )
            except Exception:
                logger.exception("Failed to import %s", item.name)

    return modules


def find_all_dataclasses() -> List[type]:
    """Find all dataclasses in whatrecord and return them as a list."""

    def should_include(obj):
        return (
            inspect.isclass(obj)
            and dataclasses.is_dataclass(obj)
            and obj not in SKIP_CLASSES
        )

    def sort_key(cls):
        return (cls.__module__, cls.__name__)

    devices = [
        obj
        for module in find_whatrecord_submodules().values()
        for _, obj in inspect.getmembers(module, predicate=should_include)
    ]

    return list(sorted(set(devices), key=sort_key))


dataclass_name_to_class = {cls.__name__: cls for cls in find_all_dataclasses()}

all_dataclasses = pytest.mark.parametrize(
    "cls", [pytest.param(cls, id=name) for name, cls in dataclass_name_to_class.items()]
)


init_args_by_type = {
    Optional[List[LoadContext]]: None,
    Optional[common.IocMetadata]: None,
    Optional[str]: None,
    Path: Path(),
    Union[int, str]: "abc",
    Union[shell.IocLoadFailure, str]: "use_cache",
    Union[str, List[str]]: ["a", "b", "c"],
    bool: True,
    bytes: b"testing",
    FullLoadContext: [LoadContext("test", 1)],
    LoadContext: LoadContext("test", 1),
    dict: {},
    float: 10.,
    int: 10,
    list: [],
    re.Pattern: re.compile("abc"),
    str: "testing",
    typing.Any: 123,
    typing.Tuple[str, str]: ("a", "b"),
    Optional[int]: 11,
    Optional[common.RecordDefinitionAndInstance]: None,
    Optional[common.RecordInstance]: None,
    Optional[common.RecordType]: None,
    Optional[common.IocshScript]: None,
    Union[snl.Declarator, snl.Variable]: snl.Variable(context=[], name="test"),
    snl.OptionalExpression: None,
    blark.dependency_store.PlcProjectMetadata: None,  # non-serialized
    blark.solution.Project: None,  # non-serialized
}


def try_to_instantiate(cls):
    kwargs = {}
    fields = dataclasses.fields(cls)
    type_hints = apischema.utils.get_type_hints(cls)
    for field in fields:
        if (
            field.default is not dataclasses.MISSING
            or field.default_factory is not dataclasses.MISSING
            or not field.init
        ):
            continue

        field_type = type_hints[field.name]
        if field_type in init_args_by_type:
            kwargs[field.name] = init_args_by_type[field_type]
        elif str(field_type) in init_args_by_type:
            kwargs[field.name] = init_args_by_type[str(field_type)]
        elif dataclasses.is_dataclass(field_type):
            kwargs[field.name] = try_to_instantiate(field_type)
        else:
            origin = apischema.typing.get_origin(field_type)
            try:
                kwargs[field.name] = init_args_by_type[origin]
            except KeyError:
                raise ValueError(f"Missing in dict: {field_type} ({origin})")

    return cls(**kwargs)


@all_dataclasses
def test_serialize(cls):
    instance = try_to_instantiate(cls)
    serialized = apischema.serialize(cls, instance)
    print(cls)
    print("Serialized:")
    print(serialized)
    deserialized = apischema.deserialize(cls, serialized)
    print("Deserialized:")
    print(deserialized)
    if cls not in SKIP_DESERIALIZATION:
        assert deserialized == instance



import asyncio
import hashlib
import json
import logging
import os
import pathlib
import sys
import textwrap
from typing import Any, Dict, Generator, List, Optional, Tuple, TypeVar, Union

import apischema

from . import settings

logger = logging.getLogger(__name__)
MODULE_PATH = pathlib.Path(__file__).parent.resolve()


T = TypeVar("T")
AnyPath = Union[str, pathlib.Path]


def get_bytes_sha256(contents: bytes):
    """Hash a byte string with the SHA-256 algorithm."""
    return hashlib.sha256(contents).hexdigest()


def check_files_up_to_date(
    file_to_hash: Dict[AnyPath, str]
) -> bool:
    """
    Check if the provided files are up-to-date by way of recorded hash vs
    current hash.

    Parameters
    ----------
    file_to_hash : Dict[Union[str, pathlib.Path], str]
        File path to hash.

    Returns
    -------
    up_to_date : bool
        If all files maintain their stored hashes, returns True.
    """
    for fn, file_hash in file_to_hash.items():
        try:
            if get_file_sha256(fn) != file_hash:
                return False
        except FileNotFoundError:
            return False

    return True


def get_file_sha256(binary: AnyPath):
    """Hash a binary with the SHA-256 algorithm."""
    # This doesn't do any sort of buffering; but our binaries are pretty small
    # in comparison to what we're storing as metadata, anyway
    with open(binary, "rb") as fp:
        return hashlib.sha256(fp.read()).hexdigest()


def read_text_file_with_hash(
    fn: pathlib.Path,
    encoding="latin-1",
) -> Tuple[str, str]:
    """Hash a binary with the SHA-256 algorithm."""
    # This doesn't do any sort of buffering; but our binaries are pretty small
    # in comparison to what we're storing as metadata, anyway
    with open(fn, "rb") as fp:
        contents = fp.read()
    sha256 = hashlib.sha256(contents).hexdigest()
    return sha256, contents.decode(encoding)


async def run_script_with_json_output(
    script_line: str,
    encoding: str = "utf-8",
    log_errors: bool = True,
) -> Optional[dict]:
    """Run a script and get its JSON output."""
    proc = await asyncio.create_subprocess_shell(
        script_line,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    (stdout, stderr) = await proc.communicate()
    if stderr and log_errors:
        stderr_text = textwrap.indent(stderr.decode("utf-8", "replace"), "    ! ")
        logger.warning(
            "Standard error output while running script (%r):\n%s",
            script_line, stderr_text
        )

    if stdout:
        return json.loads(stdout.decode(encoding))

    if log_errors:
        logger.warning(
            "No standard output while running script (%r)",
            script_line
        )


async def run_gdb(
    script: str,
    binary: Union[pathlib.Path, str],
    cls: T,
    args: Optional[List[str]] = None,
    gdb_path: Optional[str] = None,
    use_cache: bool = True,
) -> T:
    """
    Run a script and deserialize its output.

    Parameters
    ----------
    script : str
        The script name to run (whatrecord.__script__, omitting .py)

    binary : str or pathlib.Path
        The binary file to load into GDB.

    cls : type
        The dataclass type to deserialize gdb's output to.

    args : list, optional
        List of string arguments to pass to gdb.

    gdb_path : str, optional
        The path to the gdb binary.  Defaults to ``WHATRECORD_GDB_PATH``
        from the environment (``gdb``).
    """
    cache_path = pathlib.Path(settings.CACHE_PATH)
    binary_hash = get_file_sha256(binary)

    hash_filename = cache_path / f"{script}_{cls.__name__}_{binary_hash}.json"
    if use_cache:
        if not settings.CACHE_PATH or not cache_path.exists():
            use_cache = False
        else:
            try:
                with open(hash_filename, "rt") as fp:
                    json_data = json.load(fp)
                return apischema.deserialize(cls, json_data)
            except FileNotFoundError:
                ...
            except Exception as ex:
                logger.warning(
                    "Failed to load cached gdb information from disk; "
                    "re-running gdb (%s, filename=%s)",
                    ex,
                    hash_filename,
                    exc_info=True
                )

    args = " ".join(f'"{arg}"' for arg in args or [])
    script_path = MODULE_PATH / "plugins" / f"{script}.py"
    gdb_path = gdb_path or settings.GDB_PATH
    to_execute = (
        f'"{gdb_path}" '
        f"--batch-silent "
        f'--command "{script_path}" '
        f'--args "{binary}" {args}'
    )

    json_data = await run_script_with_json_output(to_execute)
    json_data = json_data or {}

    if use_cache:
        with open(hash_filename, "wt") as fp:
            json.dump(json_data, fp, indent=4)

    try:
        return apischema.deserialize(cls, json_data)
    except Exception as ex:
        ex.json_data = json_data
        raise


def find_binary_from_hashbang(
    startup_script: Optional[Union[str, pathlib.Path]],
    must_exist: bool = False,
) -> Optional[str]:
    """
    Find the binary associated with a given startup script by looking at its
    shebang.

    Returns
    -------
    binary_path : str or None
        The path to the binary, if available.
    """
    if startup_script is None:
        return None

    try:
        with open(startup_script, "rt") as fp:
            first_line = fp.read().splitlines()[0]
    except Exception:
        return None

    if first_line.startswith("#!"):
        parent_dir = pathlib.Path(startup_script).parent
        binary = parent_dir / first_line.lstrip("#!").strip()
        if not must_exist or binary.exists():
            return str(binary.resolve())


def lines_between(
    text: str,
    start_marker: str,
    end_marker: str,
    *,
    include_blank: bool = False
) -> Generator[str, None, None]:
    """
    From a block of text, yield all lines between `start_marker` and
    `end_marker`

    Parameters
    ----------
    text : str
        The block of text
    start_marker : str
        The block-starting marker to match
    end_marker : str
        The block-ending marker to match
    include_blank : bool, optional
        Skip yielding blank lines

    Yields
    ------
    line : str
        Line of text found between the markers.
    """
    found_start = False
    start_marker = start_marker.lower()
    end_marker = end_marker.lower()
    for line in text.splitlines():
        line_lowercase = line.strip().lower()
        if line_lowercase == start_marker:
            found_start = True
        elif found_start:
            if line_lowercase == end_marker:
                break
            elif line_lowercase or include_blank:
                yield line


def write_to_file(
    obj: Any,
    filename: Optional[str] = None,
    format: str = "json",
    catch_broken_pipe: bool = True
) -> str:
    """
    Write ``obj`` to ``filename`` in the specified ``format``.

    Parameters
    ----------
    obj : Any
        The object to write to the file.
    filename : Optional[str], optional
        If None, defaults to standard output.
    format : str, optional
        The format to use.  "json" is supported for all types. Other types
        may only support "console" or "file".

    Returns
    -------
    data : str
        The data written to the file.
    """
    if format == "json":
        json_info = apischema.serialize(obj)
        to_write = json.dumps(json_info, indent=settings.INDENT)
    else:
        from .format import FormatContext
        fmt = FormatContext()
        to_write = fmt.render_object(obj, format)

    if filename is None:
        try:
            print(to_write, flush=True)
        except BrokenPipeError:
            if not catch_broken_pipe:
                raise

            # Ref: https://docs.python.org/3/library/signal.html#note-on-sigpipe
            # Python flushes standard streams on exit; redirect remaining
            # output to devnull to avoid another BrokenPipeError at shutdown
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            sys.exit(1)  # Python exits with error code 1 on EPIPE
    else:
        with open(filename, "wt") as fp:
            fp.write(to_write)

    return to_write



"""
"whatrecord info" is used to get PV information from the whatrecord server.
"""

import argparse
import json
import logging
import sys

import apischema

from ..client import get_record_info
from ..format import FormatContext

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        "records", type=str, nargs="+", help="Record name(s)"
    )

    parser.add_argument(
        "--json", dest="as_json", action="store_true", help="Output raw JSON"
    )
    return parser


async def main(records, as_json=False, file=sys.stdout):
    info = await get_record_info(*records)
    if as_json:
        # TODO: JSON -> obj -> JSON round tripping
        json_info = apischema.serialize(info)
        print(json.dumps(json_info, indent=4))
    else:
        fmt = FormatContext()
        for pv, pv_get_info in info.items():
            print(pv, file=file)
            print("-" * len(pv), file=file)
            print(file=file)
            print(fmt.render_object(pv_get_info, "console"), file=file)



from __future__ import annotations

import asyncio
import json
import logging
import os
import pathlib
import signal
import sys
import textwrap
import traceback
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass, field
from typing import (Any, ClassVar, Dict, Generator, Iterable, List, Optional,
                    Tuple, Union)

import apischema
import lark
from epicsmacrolib import IocshRedirect

from . import common, dbtemplate, graph, settings, util
from .access_security import AccessSecurityState
from .asyn import AsynState
from .autosave import AutosaveState
from .common import (AnyPath, FullLoadContext, IocMetadata, IocshCmdArgs,
                     IocshResult, IocshScript, LoadContext, MutableLoadContext,
                     PVRelations, RecordDefinitionAndInstance, RecordInstance,
                     ShellStateHandler, StringWithContext, WhatRecord,
                     time_context)
from .db import Database, DatabaseLoadFailure, RecordType
from .format import FormatContext
from .iocsh import parse_iocsh_line
from .macro import MacroContext
from .motor import MotorState
from .streamdevice import StreamDeviceState

logger = logging.getLogger(__name__)


_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class ShellState(ShellStateHandler):
    """
    IOC shell state container.

    Contains hooks for commands and state information.

    This base state handler should only handle epics base-defined IOC shell
    commands, including: paths, variables, database loading, and IOC
    initialization.

    It is the top-level state container, which sub handlers should rely on for
    things like loading files and other core state information.

    Attributes
    ----------
    prompt : str
        The prompt - PS1 - as in "epics>".
    variables : dict
        Shell variables (not environment variables).
    string_encoding : str
        String encoding for byte strings and files.
    macro_context : MacroContext
        Macro context for commands that are evaluated.
    standin_directories : dict
        Rewrite hard-coded directory prefixes by setting::

            standin_directories = {"/replace_this/": "/with/this"}
    loaded_files : Dict[str, str]
        Files loaded, mapped to a hash of their contents.
    working_directory : pathlib.Path
        Current working directory.
    database_definition : Database
        Loaded database definition (dbd).
    database : Dict[str, RecordInstance]
        The IOC database of records.
    pva_database : Dict[str, RecordInstance]
        The IOC database of PVAccess groups.
    aliases : Dict[str, str]
        Alias name to record name.
    load_context : List[MutableLoadContext]
        Current loading context stack (e.g., ``st.cmd`` then
        ``common_startup.cmd``).  Modified in place as scripts are evaluated.
    """

    prompt: str = "epics>"
    variables: Dict[str, str] = field(default_factory=dict)
    string_encoding: str = "latin-1"
    ioc_initialized: bool = False
    standin_directories: Dict[str, str] = field(default_factory=dict)
    working_directory: pathlib.Path = field(
        default_factory=lambda: pathlib.Path.cwd(),
    )
    aliases: Dict[str, str] = field(default_factory=dict)
    database_definition: Optional[Database] = None
    database: Dict[str, RecordInstance] = field(default_factory=dict)
    pva_database: Dict[str, RecordInstance] = field(default_factory=dict)
    load_context: List[MutableLoadContext] = field(default_factory=list)
    loaded_files: Dict[str, str] = field(default_factory=dict)
    macro_context: MacroContext = field(default_factory=MacroContext)
    ioc_info: IocMetadata = field(default_factory=IocMetadata)
    db_add_paths: List[pathlib.Path] = field(default_factory=list)
    database_grammar_version: int = 3

    # Sub-state handlers:
    access_security: AccessSecurityState = field(default_factory=AccessSecurityState)
    asyn: AsynState = field(default_factory=AsynState)
    autosave: AutosaveState = field(default_factory=AutosaveState)
    motor: MotorState = field(default_factory=MotorState)
    streamdevice: StreamDeviceState = field(default_factory=StreamDeviceState)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "console": textwrap.dedent(
            """\
            {{ obj | classname }}:
            """.rstrip(),
        ),
        "console-verbose": textwrap.dedent(
            """\
            {{ obj | classname }}:
            """.rstrip(),
        )
    }

    def __post_init__(self):
        super().__post_init__()
        self.macro_context.string_encoding = self.string_encoding

    @property
    def sub_handlers(self) -> List[ShellStateHandler]:
        """Handlers which contain their own state."""
        return [
            self.access_security,
            self.asyn,
            self.autosave,
            self.motor,
            self.streamdevice,
        ]

    def load_file(self, filename: AnyPath) -> Tuple[pathlib.Path, str]:
        """Load a file, record its hash, and return its contents."""
        filename = self._fix_path(filename)
        filename = filename.resolve()
        shasum, contents = util.read_text_file_with_hash(
            filename, encoding=self.string_encoding
        )
        self.loaded_files[str(filename)] = shasum
        self.ioc_info.loaded_files[str(filename)] = shasum
        return filename, contents

    def _handle_input_redirect(
        self,
        redir: IocshRedirect,
        shresult: IocshResult,
        recurse: bool = True,
        raise_on_error: bool = False,
    ):
        try:
            filename, contents = self.load_file(redir.name)
        except Exception as ex:
            shresult.error = f"{type(ex).__name__}: {redir.name}"
            yield shresult
            return

        yield shresult
        yield from self.interpret_shell_script_text(
            contents.splitlines(), recurse=recurse, name=filename
        )

    def interpret_shell_line(self, line, recurse=True, raise_on_error=False):
        """Interpret a single shell script line."""
        shresult = parse_iocsh_line(
            line,
            context=self.get_load_context(),
            prompt=self.prompt,
            macro_context=self.macro_context,
            string_encoding=self.string_encoding,
        )
        input_redirects = [redir for redir in shresult.redirects if redir.mode == "r"]
        if shresult.error:
            yield shresult
        elif input_redirects:
            if recurse:
                yield from self._handle_input_redirect(
                    input_redirects[0],
                    shresult,
                    recurse=recurse,
                    raise_on_error=raise_on_error,
                )
        elif shresult.argv:
            try:
                result = self._handle_command(*shresult.argv)
                if result:
                    # Only set if not-None to speed up serialization
                    shresult.result = result
            except Exception as ex:
                if raise_on_error:
                    raise
                ex_details = traceback.format_exc()
                shresult.error = f"Failed to execute: {ex}:\n{ex_details}"

            yield shresult
            if isinstance(shresult.result, IocshCmdArgs):
                yield from self.interpret_shell_line(
                    shresult.result.command, recurse=recurse
                )
        else:
            # Otherwise, nothing to do
            yield shresult

    def interpret_shell_script(
        self,
        filename: Union[pathlib.Path, str],
        recurse: bool = True,
        raise_on_error: bool = False,
    ) -> Generator[IocshResult, None, None]:
        """Load and interpret a shell script named ``filename``."""
        filename, contents = self.load_file(filename)
        yield from self.interpret_shell_script_text(
            contents.splitlines(),
            name=str(filename),
            recurse=recurse,
            raise_on_error=raise_on_error,
        )

    def interpret_shell_script_text(
        self,
        lines: Iterable[str],
        name: str = "unknown",
        recurse: bool = True,
        raise_on_error: bool = False,
    ) -> Generator[IocshResult, None, None]:
        """Interpret a shell script named ``name`` with ``lines`` of text."""
        load_ctx = MutableLoadContext(str(name), 0)
        try:
            self.load_context.append(load_ctx)
            for lineno, line in enumerate(lines, 1):
                load_ctx.line = lineno
                yield from self.interpret_shell_line(
                    line,
                    recurse=recurse,
                    raise_on_error=raise_on_error,
                )
        finally:
            self.load_context.remove(load_ctx)

        # for rec in list(self.database.values()) + list(self.pva_database.values()):
        #     try:
        #         rec.metadata.update(self.annotate_record(rec))
        #     except Exception:
        #         logger.exception("Failed to annotate record: %s", rec.name)

    def get_load_context(self) -> FullLoadContext:
        """Get a FullLoadContext tuple representing where we are now."""
        if not self.load_context:
            return tuple()
        return tuple(ctx.to_load_context() for ctx in self.load_context)

    def _handle_command(self, command, *args):
        """Handle IOC shell 'command' with provided arguments."""
        handler = self._handlers.get(command, None)
        if handler is not None:
            return handler(*args)
        return self.unhandled(command, args)

    def _fix_path(self, filename: AnyPath) -> pathlib.Path:
        """
        Makes filename an absolute path with respect to the working directory.

        Also replaces standin directories, if provided an absolute path.
        """
        filename = str(filename)
        if os.path.isabs(filename):
            for from_, to in self.standin_directories.items():
                if filename.startswith(from_):
                    _, suffix = filename.split(from_, 1)
                    return pathlib.Path(to + suffix)

        return self.working_directory / filename

    @property
    def db_include_paths(self) -> List[pathlib.Path]:
        """Database include paths (EPICS_DB_INCLUDE_PATH)."""
        env_paths = self.paths_from_env_var("EPICS_DB_INCLUDE_PATH")
        if not env_paths:
            return [self.working_directory] + self.db_add_paths
        return env_paths + self.db_add_paths

    def paths_from_env_var(
        self,
        env_var: str,
        *,
        default: Optional[str] = None
    ) -> List[pathlib.Path]:
        """Paths from an environment variable (or macro)."""
        env_var = self.macro_context.get(env_var, default) or ""
        return [
            (self.working_directory / pathlib.Path(path)).resolve()
            # TODO: this is actually OS-dependent (: on linux, ; on Windows)
            for path in env_var.split(":")
        ]

    def _fix_path_with_search_list(
        self,
        filename: Union[str, pathlib.Path],
        include_paths: List[pathlib.Path],
    ) -> pathlib.Path:
        """Given a list of paths, find ``filename``."""
        filename = str(filename)
        if not include_paths or str(filename)[0] in ("/", "\\"):
            # Include path unset or even something resembling a nested path
            # automatically is used as-is
            return self._fix_path(self.working_directory / filename)

        for path in include_paths:
            option = self._fix_path(path) / filename
            if option.exists() and option.is_file():
                return option

        if "/" in filename or "\\" in filename:
            return self._fix_path(self.working_directory / filename)

        paths = list(str(path) for path in include_paths)
        raise FileNotFoundError(
            f"File {filename!r} not found in search path: {paths} "
        )

    def unhandled(self, command, args):
        ...
        # return f"No handler for handle_{command}"

    @_handler
    def handle_iocshRegisterVariable(self, variable: str, value: str = ""):
        self.variables[variable] = value
        return f"Registered variable: {variable!r}={value!r}"

    def env_set_EPICS_BASE(self, path):
        # TODO: slac-specific
        path = str(pathlib.Path(path).resolve())
        version_prefixes = [
            "/reg/g/pcds/epics/base/",
            "/cds/group/pcds/epics/base/",
        ]
        for prefix in version_prefixes:
            if path.startswith(prefix):
                path = path[len(prefix):]
                if "/" in path:
                    path = path.split("/")[0]
                version = path.lstrip("R")
                if self.ioc_info.base_version == settings.DEFAULT_BASE_VERSION:
                    self.ioc_info.base_version = version
                    return f"Set base version: {version}"
                return (
                    f"Found version ({version}) but version already specified:"
                    f" {self.ioc_info.base_version}"
                )

    @_handler
    def handle_epicsEnvSet(self, variable: str, value: str = ""):
        self.macro_context.define(**{variable: value})
        hook = getattr(self, f"env_set_{variable}", None)
        if hook and callable(hook):
            hook_result = hook(value)
            if hook_result:
                return {
                    "hook": hook_result,
                }

    @_handler
    def handle_epicsEnvShow(self):
        return self.macro_context.get_macros()

    def handle_iocshCmd(self, command: str = "", *_):
        # TODO: odd return type, used below
        return IocshCmdArgs(context=self.get_load_context(), command=command)

    @_handler
    def handle_cd(self, path: str = ""):
        if not path:
            raise RuntimeError("Invalid directory path, ignored")

        path = self._fix_path(path)
        if path.is_absolute():
            new_dir = path
        else:
            new_dir = self.working_directory / path

        if not new_dir.exists():
            raise RuntimeError(f"Path does not exist: {new_dir}")

        self.working_directory = new_dir.resolve()
        os.environ["PWD"] = str(self.working_directory)
        return {
            "result": f"New working directory: {self.working_directory}"
        }

    handle_chdir = handle_cd

    @_handler
    def handle_iocInit(self):
        if self.ioc_initialized:
            return {
                "success": False,
                "error": "Already initialized",
            }

        result = {
            "success": True,
        }

        for handler in self.sub_handlers:
            handler_result = handler.pre_ioc_init()
            result.update(handler_result or {})

        self.ioc_initialized = True

        for handler in self.sub_handlers:
            handler_result = handler.post_ioc_init()
            result.update(handler_result or {})

        return result

    @_handler
    def handle_dbLoadDatabase(self, dbd: str, path: str = "", substitutions: str = ""):
        if self.ioc_initialized:
            raise RuntimeError("Database cannot be loaded after iocInit")
        if self.database_definition:
            # TODO: technically this is allowed; we'll need to update
            # raise RuntimeError("dbd already loaded")
            return {
                "error": "whatrecord: TODO multiple dbLoadDatabase"
            }

        dbd_path = self._fix_path_with_search_list(dbd, self.db_include_paths)
        fn, contents = self.load_file(dbd_path)
        macro_context = MacroContext(use_environment=False)
        macro_context.define_from_string(substitutions or "")

        self.database_grammar_version = self.ioc_info.database_version_spec
        try:
            self.database_definition = Database.from_string(
                contents,
                version=self.database_grammar_version,
                filename=fn,
                macro_context=macro_context,
            )
        except lark.exceptions.ParseError as original_ex:
            # Our guess about the EPICS version may have been wrong.
            # Let's see if the dbd is valid for the other version.
            other_spec = {
                3: 4,
                4: 3
            }[self.database_grammar_version]

            try:
                self.database_definition = Database.from_string(
                    contents,
                    version=other_spec,
                    filename=fn,
                    macro_context=macro_context,
                )
            except Exception:
                raise DatabaseLoadFailure(
                    f"Database definition file failed to load. "
                    f"Attempted to load with grammar version {self.database_grammar_version}, "
                    f"and fallback version {other_spec} unsuccessfully."
                ) from original_ex

            # Our initial guess was wrong... Load further database files with
            # the adjusted version.
            self.database_grammar_version = other_spec

        for addpath in self.database_definition.addpaths:
            for path in addpath.path.split(os.pathsep):  # TODO: OS-dependent
                self.db_add_paths.append((dbd_path.parent / path).resolve())

        self.aliases.update(self.database_definition.aliases)
        return {
            "context": [LoadContext(str(fn), 0)],
            "result": "Loaded database",
            "grammar_version": self.database_grammar_version,
            "record_types": list(sorted(self.database_definition.record_types)),
            "drivers": list(sorted(self.database_definition.drivers)),
            "functions": list(sorted(self.database_definition.functions)),
            "breaktables": list(sorted(self.database_definition.breaktables)),
        }

    @_handler
    def handle_dbLoadTemplate(self, filename: str, macros: str = ""):
        filename = self._fix_path_with_search_list(filename, self.db_include_paths)
        filename, contents = self.load_file(filename)

        # TODO this should be multiple load calls for the purposes of context
        result = {
            "context": [LoadContext(str(filename), 0)],
            "total_records": 0,
            "total_groups": 0,
            "loaded_files": [],
        }

        template = dbtemplate.TemplateSubstitution.from_string(contents, filename=filename)
        for sub in template.substitutions:
            database_contents = sub.expand_file(search_paths=self.db_include_paths)
            # TODO loading file twice (ensure it gets added to the loaded_files list)
            self.load_file(sub.filename)
            db = self._load_database(
                filename=str(sub.filename),
                contents=database_contents,
                macros=macros,
                context=self.get_load_context() + sub.context,
            )
            info = {
                "context": [LoadContext(str(sub.filename), 0)],
                "macros": sub.macros,
                "records": len(db.records),
                "groups": len(db.pva_groups),
                "lint": db.lint,
            }
            result["total_records"] += len(db.records)
            result["total_groups"] += len(db.pva_groups)
            result["loaded_files"].append(info)

        return result

    def _load_database(
        self,
        filename: str,
        contents: str,
        macros: str,
        context: FullLoadContext
    ) -> Database:
        macro_context = MacroContext(use_environment=False)
        macros = macro_context.define_from_string(macros or "")

        try:
            db = Database.from_string(
                contents,
                filename=filename,
                dbd=self.database_definition,
                macro_context=macro_context,
                version=self.database_grammar_version,
            )
        except Exception as ex:
            # TODO move this around
            raise DatabaseLoadFailure(
                f"Failed to load {filename}: {type(ex).__name__} {ex}"
            ) from ex

        for name, rec in db.records.items():
            if name not in self.database:
                self.database[name] = rec
                rec.context = context + rec.context
                rec.owner = self.ioc_info.name
            else:
                entry = self.database[name]
                entry.context = entry.context + rec.context
                entry.fields.update(rec.fields)
                # entry.owner = self.ioc_info.name ?

        for name, rec in db.pva_groups.items():
            if name not in self.pva_database:
                self.pva_database[name] = rec
                rec.context = context + rec.context
                rec.owner = self.ioc_info.name
            else:
                entry = self.database[name]
                entry.context = entry.context + rec.context
                entry.fields.update(rec.fields)
                # entry.owner = self.ioc_info.name ?

        self.aliases.update(db.aliases)
        for addpath in db.addpaths:
            for path in addpath.path.split(os.pathsep):  # TODO: OS-dependent
                self.db_add_paths.append((db.parent / path).resolve())

        return db

    @_handler
    def handle_dbLoadRecords(self, filename: str, macros: str = ""):
        if not self.database_definition:
            raise RuntimeError("dbd not yet loaded")
        if self.ioc_initialized:
            raise RuntimeError("Records cannot be loaded after iocInit")

        filename = self._fix_path_with_search_list(filename, self.db_include_paths)
        filename, contents = self.load_file(filename)
        db = self._load_database(
            filename=filename,
            contents=contents,
            macros=macros or "",
            context=self.get_load_context()
        )

        for record in db.records.values():
            # TODO: maybe add option to annotate all records without requiring
            # per-record whatrecord queries
            # if self.annotate_all:
            #       record.metadata.update(self.annotate_record(record))

            # StreamDevice loading is context-sensitive: the current working
            # directory at the time of evaluation may be used to find
            # the protocol file.  Special case this here and pre-annotate.
            annotation = self.streamdevice.annotate_record(record)
            if annotation:
                key = StringWithContext(
                    self.streamdevice.metadata_key,
                    context=(),
                )
                record.metadata[key] = annotation

        return {
            "context": [LoadContext(str(filename), 0)],
            "num_records": len(db.records),
            "num_pva_groups": len(db.pva_groups),
            "lint": db.lint,
        }

    def annotate_record(self, record: RecordInstance) -> Dict[str, Any]:
        """Hook to annotate a record after being loaded."""
        result = {}
        for handler in self.sub_handlers:
            if handler is self.streamdevice:
                # This is handled as the database is loaded.
                continue

            try:
                annotation = handler.annotate_record(record)
            except Exception:
                logger.exception(
                    "Record annotation failed for %s with handler %s",
                    record.name, type(handler).__name__
                )
            else:
                if annotation is not None:
                    # record.metadata[handler.metadata_key] = annotation
                    result[handler.metadata_key] = annotation

        return result

    @_handler
    def handle_dbl(self, rtyp: str = "", fields: str = ""):
        ...

    @_handler
    def handle_NDPvaConfigure(
        self,
        portName: str,
        queueSize: int = 0,
        blockingCallbacks: str = "",
        NDArrayPort: str = "",
        NDArrayAddr: str = "",
        pvName: str = "",
        maxBuffers: int = 0,
        maxMemory: int = 0,
        priority: int = 0,
        stackSize: int = 0,
    ):
        """Implicitly creates a PVA group named ``pvName``."""
        metadata = {
            "portName": portName or "",
            "queueSize": queueSize or "",
            "blockingCallbacks": blockingCallbacks or "",
            "NDArrayPort": NDArrayPort or "",
            "NDArrayAddr": NDArrayAddr or "",
            "pvName": pvName or "",
            "maxBuffers": maxBuffers or "",
            "maxMemory": maxMemory or "",
            "priority": priority or "",
            "stackSize": stackSize or "",
        }
        self.pva_database[pvName] = RecordInstance(
            context=self.get_load_context(),
            name=pvName,
            record_type="PVA",
            fields={},
            is_pva=True,
            metadata={"areaDetector": metadata},
        )
        return metadata


@dataclass
class ScriptContainer:
    """
    Aggregate container for any number of LoadedIoc instances.

    Combines databases, sets of loaded files ease of querying.
    """

    database: Dict[str, RecordInstance] = field(default_factory=dict)
    aliases: Dict[str, str] = field(default_factory=dict)
    pva_database: Dict[str, RecordInstance] = field(default_factory=dict)
    scripts: Dict[str, LoadedIoc] = field(default_factory=dict)
    startup_script_to_ioc: Dict[str, str] = field(default_factory=dict)
    #: absolute filename path to sha
    loaded_files: Dict[str, str] = field(default_factory=dict)
    record_types: Dict[str, RecordType] = field(default_factory=dict)
    pv_relations: PVRelations = field(default_factory=dict)

    def add_loaded_ioc(self, loaded: LoadedIoc):
        ioc_name = loaded.metadata.name
        self.scripts[ioc_name] = loaded
        self.startup_script_to_ioc[str(loaded.metadata.script)] = ioc_name

        # TODO: IOCs will have conflicting definitions of records
        self.aliases.update(loaded.shell_state.aliases)
        if loaded.shell_state.database_definition:
            self.record_types.update(
                loaded.shell_state.database_definition.record_types
            )
        graph.combine_relations(
            self.pv_relations, self.database,
            loaded.pv_relations, loaded.shell_state.database,
            record_types=self.record_types,
            aliases=self.aliases,
        )
        self.database.update(loaded.shell_state.database)
        self.pva_database.update(loaded.shell_state.pva_database)
        self.loaded_files.update(loaded.shell_state.loaded_files)

    def whatrec(
        self,
        rec: str,
        field: Optional[str] = None,
        include_pva: bool = True,
        format_option: str = "console",
        file=sys.stdout,
    ) -> List[WhatRecord]:
        fmt = FormatContext()
        result = []
        for name, loaded in self.scripts.items():
            info: Optional[WhatRecord] = loaded.whatrec(
                rec,
                field,
                include_pva=include_pva,
            )
            if info is not None:
                info.ioc = loaded.metadata
                for match in [info.record, info.pva_group]:
                    if file is not None and match is not None:
                        print(fmt.render_object(match, format_option), file=file)
                result.append(info)
        return result


@dataclass
class LoadedIoc:
    name: str
    path: pathlib.Path
    metadata: IocMetadata
    shell_state: ShellState
    script: IocshScript
    load_failure: bool = False
    pv_relations: PVRelations = field(default_factory=dict)

    _jinja_format_: ClassVar[Dict[str, str]] = {
        "file": textwrap.dedent(
            """\
            {{ render_object(script, "file") }}
            """.rstrip(),
        ),
        "console": textwrap.dedent(
            """\
            {{ obj | classname }}:
            path: {{ path }}
            metadata: {% set metadata = render_object(metadata, "console") %}
            {{ metadata | indent(4) }}
            shell_state: {% set shell_state = render_object(shell_state, "console") %}
            {{ shell_state | indent(4) }}
            script:
            {% set script = render_object(script, "console") %}
                {{ script | indent(4) }}
            load_failure: {{ load_failure }}
            """.rstrip(),
        ),
        "console-verbose": textwrap.dedent(
            """\
            {{ obj | classname }}:
            path: {{ path }}
            metadata: {% set metadata = render_object(metadata, "console-verbose") %}
            {{ metadata | indent(4) }}
            shell_state: {% set shell_state = render_object(shell_state, "console-verbose") %}
            {{ shell_state | indent(4) }}
            script:
            {% set script = render_object(script, "console-verbose") %}
            {{ script | indent(4) }}
            load_failure: {{ load_failure }}
            pv_relations: {{ pv_relations }}
            """.rstrip(),
        )
    }

    @classmethod
    def _json_from_cache(cls, md: IocMetadata) -> Optional[dict]:
        """
        Load the serialized JSON from the cache for the given IOC.

        Parameters
        ----------
        md : IocMetadata
            The metadata that identifies the IOC.

        Returns
        -------
        dict or None
            The cached JSON, if available.
        """
        try:
            with open(md.ioc_cache_filename, "rb") as fp:
                return json.load(fp)
        except FileNotFoundError:
            ...
        except json.decoder.JSONDecodeError:
            # Truncated output file, perhaps
            ...

    @classmethod
    def from_cache(cls, md: IocMetadata) -> Optional[LoadedIoc]:
        """
        If available, load the cached data about the provided IOC.

        Parameters
        ----------
        md : IocMetadata
            The metadata that identifies the IOC.

        Returns
        -------
        LoadedIoc or None
        """
        json_dict = cls._json_from_cache(md)
        if json_dict is not None:
            return apischema.deserialize(cls, json_dict)

    def save_to_cache(self) -> bool:
        """Save this instance to the application cache."""
        if not settings.CACHE_PATH:
            return False

        with open(self.metadata.ioc_cache_filename, "wt") as fp:
            json.dump(apischema.serialize(self), fp=fp)
        return True

    @classmethod
    def from_errored_load(
        cls, md: IocMetadata, load_failure: IocLoadFailure
    ) -> LoadedIoc:
        """
        Create a LoadedIoc from a failed load, as indicated by IocLoadFailure.

        Parameters
        ----------
        md : IocMetadata
            The metadata that was used to load the IOC.
        load_failure : IocLoadFailure
            The load failure information.

        Returns
        -------
        LoadedIoc
        """
        exception_line = f"{load_failure.ex_class}: {load_failure.ex_message}"
        error_lines = [exception_line] + load_failure.traceback.splitlines()
        script = IocshScript(
            path=str(md.script),
            lines=tuple(
                IocshResult(line=line, context=(LoadContext("error", lineno),))
                for lineno, line in enumerate(error_lines, 1)
            ),
        )
        md.metadata["exception_class"] = load_failure.ex_class
        md.metadata["exception_message"] = load_failure.ex_message
        md.metadata["traceback"] = load_failure.traceback
        md.metadata["load_failure"] = True
        return cls(
            name=md.name,
            path=md.script,
            metadata=md,
            shell_state=ShellState(),
            script=script,
            load_failure=True,
        )

    @classmethod
    def from_metadata(cls, md: IocMetadata) -> LoadedIoc:
        """
        Generate a LoadedIoc by interpreting the startup script.

        Parameters
        ----------
        md : IocMetadata
            The metadata from the IOC.

        Returns
        -------
        LoadedIoc
        """
        sh = ShellState(ioc_info=md)
        sh.working_directory = md.startup_directory
        sh.macro_context.define(**md.macros)
        sh.standin_directories = md.standin_directories or {}

        # It's not enough to chdir, as we can rely on the environment variable
        # in shell scripts:
        os.environ["PWD"] = str(md.startup_directory)

        script = IocshScript.from_metadata(md, sh=sh)
        return cls(
            name=md.name,
            path=md.script,
            metadata=md,
            shell_state=sh,
            script=script,
            pv_relations=graph.build_database_relations(sh.database),
        )

    def whatrec(
        self, rec: str, field: Optional[str] = None, include_pva: bool = True
    ) -> Optional[WhatRecord]:
        """
        Get record information, optionally including PVAccess results.

        Parameters
        ----------
        rec : str
            The record name (excluding any field).
        field : Optional[str], optional
            An optional field name, such as ``DESC``.  Currently unused.
        include_pva : bool, optional
            Include PVAccess results in the query.

        Returns
        -------
        Optional[WhatRecord]
            Information about the provided record, if available.
        """
        state = self.shell_state
        v3_inst = state.database.get(state.aliases.get(rec, rec), None)
        pva_inst = state.pva_database.get(rec, None) if include_pva else None
        if not v3_inst and not pva_inst:
            return

        what = WhatRecord(
            name=rec,
            ioc=self.metadata,
            record=None,
            pva_group=None,
        )
        if v3_inst is not None:
            if not state.database_definition:
                defn = None
            else:
                defn = state.database_definition.record_types.get(
                    v3_inst.record_type, None
                )
                what.menus = state.database_definition.menus
                # but what about device types and such?

            v3_inst.metadata.update(state.annotate_record(v3_inst))
            what.record = RecordDefinitionAndInstance(defn, v3_inst)

        if pva_inst is not None:
            what.pva_group = pva_inst

        return what


def load_cached_ioc(
    md: IocMetadata,
    allow_failed_load: bool = False,
) -> Optional[LoadedIoc]:
    """
    Load an IOC from the cache.

    Parameters
    ----------
    md : IocMetadata
        The metadata from the IOC, which determines its cache key.
    allow_failed_load : bool, optional
        A previous load of the provided IOC may have been marked as a failure.
        If set, ``load_cached_ioc`` will return the results of the failed load.
        If unset, ``load_cached_ioc`` will return ``None`` and allow the caller
        to treat this as a cache miss and potentially reload it from disk.

    Returns
    -------
    Optional[LoadedIoc]
        The cached LoadedIoc instance.
    """
    cached_md = md.from_cache()
    if cached_md is None:
        logger.debug("Cached metadata unavailable %s", md.name)
        return None

    if md._cache_key != cached_md._cache_key:
        logger.error("Cache key mismatch?! %s %s", md._cache_key, cached_md._cache_key)
        return None

    if allow_failed_load and (
        cached_md.metadata.get("load_failure") or md.looks_like_sh
    ):
        logger.debug(
            "%s allow_failed_load=True; %s, md.looks_like_sh=%s",
            md.name,
            cached_md.metadata.get("load_failure"),
            md.looks_like_sh,
        )
    elif not cached_md.is_up_to_date():
        logger.debug("%s is not up-to-date", md.name)
        return

    try:
        logger.debug("%s is up-to-date; load from cache", md.name)
        return LoadedIoc._json_from_cache(cached_md)
    except FileNotFoundError:
        logger.error("%s is noted as up-to-date; but cache file missing", md.name)

    return None


@dataclass
class IocLoadFailure:
    ex_class: str
    ex_message: str
    traceback: str


@dataclass
class IocLoadResult:
    identifier: Union[int, str]
    load_time: float
    cache_hit: bool
    result: Union[IocLoadFailure, str]


async def async_load_ioc(
    identifier: Union[int, str],
    md: IocMetadata,
    standin_directories,
    use_gdb: bool = True,
    use_cache: bool = True,
) -> IocLoadResult:
    """
    Helper function for loading an IOC in a subprocess and relying on the cache.

    Parameters
    ----------
    standin_directories :
        Stand-in directories to use when loading the IOC.
    identifier : Union[int, str]
        The IOC identifier.
    md : IocMetadata
        Metadata tied to the IOC.
    use_gdb : bool, optional
        Use GDB to analyze the IOC binary for linting of shell script commands.
    use_cache : bool, optional
        Load from the cache for IOCs which have been cached.

    Returns
    -------
    IocLoadResult
        The result of loading, including an identifier which can be used by the
        caller to load the item from disk (as it's faster than using
        multiprocessing).
    """
    if not settings.CACHE_PATH:
        use_cache = False
    with time_context() as ctx:
        try:
            md.standin_directories.update(standin_directories)
            if use_cache:
                cached_ioc = load_cached_ioc(md)
                if cached_ioc:
                    return IocLoadResult(
                        identifier=identifier,
                        load_time=ctx(),
                        cache_hit=True,
                        result="use_cache"
                    )

            loaded = LoadedIoc.from_metadata(md)
            if use_gdb:
                await md.get_binary_information()

            if use_cache:
                loaded.metadata.save_to_cache()
                loaded.save_to_cache()
                # Avoid pickling massive JSON blob; instruct server to load
                # from cache with token 'use_cache'
                serialized = "use_cache"
            else:
                serialized = apischema.serialize(loaded)
        except Exception as ex:
            return IocLoadResult(
                identifier=identifier,
                load_time=ctx(),
                cache_hit=False,
                result=IocLoadFailure(
                    ex_class=type(ex).__name__,
                    ex_message=str(ex),
                    traceback=traceback.format_exc(),
                ),
            )

        return IocLoadResult(
            identifier=identifier,
            load_time=ctx(),
            cache_hit=False,
            result=serialized,
        )


def _load_ioc(
    identifier: Union[int, str],
    md: IocMetadata,
    standin_directories,
    use_gdb: bool = True,
    use_cache: bool = True,
) -> IocLoadResult:
    """A synchronous wrapper around ``async_load_ioc``."""
    return asyncio.run(
        async_load_ioc(
            identifier=identifier, md=md,
            standin_directories=standin_directories, use_gdb=use_gdb,
            use_cache=use_cache
        )
    )


def _sigint_handler(signum, frame):
    """SIGINT handler for subprocesses."""
    logger.error("Subprocess killed with SIGINT; exiting.")
    sys.exit(1)


def _process_init():
    """Subprocess initializer; enables SIGINT handler."""
    signal.signal(signal.SIGINT, _sigint_handler)


async def load_startup_scripts_with_metadata(
    *md_items,
    standin_directories=None,
    processes: int = 8,
    use_gdb: bool = True,
) -> ScriptContainer:
    """
    Load all given startup scripts into a shared ScriptContainer.

    Parameters
    ----------
    *md_items : list of IocMetadata
        List of IOC metadata.
    standin_directories : dict
        Stand-in/substitute directory mapping.
    processes : int
        The number of processes to use when loading.
    """
    total_files = len(md_items)
    total_child_load_time = 0.0

    with time_context() as total_time, ProcessPoolExecutor(
        max_workers=processes, initializer=_process_init
    ) as executor:
        coros = [
            asyncio.wrap_future(
                executor.submit(
                    _load_ioc, identifier=idx, md=md,
                    standin_directories=standin_directories, use_gdb=use_gdb
                )
            )
            for idx, md in enumerate(md_items)
        ]

        for coro in asyncio.as_completed(coros):
            try:
                load_result = await coro
                md = md_items[load_result.identifier]
            except Exception as ex:
                logger.exception(
                    "Internal error while loading: %s: %s [server %.1f s]",
                    type(ex).__name__,
                    ex,
                    total_time(),
                )
                continue

            use_cache = load_result.result == "use_cache"
            if not use_cache:
                loaded = load_result.result
            else:
                try:
                    loaded = load_cached_ioc(md, allow_failed_load=True)
                    if loaded is None:
                        raise ValueError("Cache entry is empty?")
                except Exception as ex:
                    logger.exception(
                        "Internal error while loading cached IOC from disk: "
                        "%s: %s [server %.1f s]",
                        type(ex).__name__,
                        ex,
                        total_time(),
                    )
                    continue

            total_child_load_time += load_result.load_time

            if isinstance(loaded, IocLoadFailure):
                failure_result: IocLoadFailure = loaded
                logger.error(
                    "Failed to load %s in subprocess: %s "
                    "[%.1f s; server %.1f]: %s\n%s",
                    md.name or md.script,
                    failure_result.ex_class,
                    load_result.load_time,
                    total_time(),
                    failure_result.ex_message,
                    (
                        failure_result.traceback
                        if failure_result.ex_class != "FileNotFoundError"
                        else ""
                    ),
                )
                if md.base_version == settings.DEFAULT_BASE_VERSION:
                    md.base_version = "unknown"
                yield md, LoadedIoc.from_errored_load(md, loaded)
                continue

            with time_context() as ctx:
                loaded_ioc = apischema.deserialize(LoadedIoc, loaded)
                logger.info(
                    "Child loaded %s%s in %.1f s, server deserialized in %.1f s",
                    md.name or md.script,
                    " from cache" if load_result.cache_hit else "",
                    load_result.load_time,
                    ctx(),
                )
                yield md, loaded_ioc

    logger.info(
        "Loaded %d startup scripts in %.1f s (wall time) with %d process(es)",
        total_files,
        total_time(),
        processes,
    )
    logger.info(
        "Child processes reported taking a total of %.1f "
        "sec, the total time on %d process(es)",
        total_child_load_time,
        processes,
    )


# TODO: apischema skip still requires forwardref to exist?
common.ShellState = ShellState



from typing import List

import apischema
from graphql import graphql_sync, print_schema

from .common import AsynPortBase, WhatRecord
from .shell import ScriptContainer


class QueryHelper:
    def __init__(self, container: ScriptContainer):
        self.container = container
        self.schema = apischema.graphql.graphql_schema(
            query=[self.whatrec, self.whatasyn]
        )

    def describe_schema(self) -> str:
        # TODO: remove
        return print_schema(self.schema)

    def whatrec(self, name: str) -> List[WhatRecord]:
        return self.container.whatrec(name, file=None)

    def whatasyn(self, name: str) -> List[AsynPortBase]:
        return []

    def query(self, query):
        return graphql_sync(self.schema, query)



import os

import apischema
import pytest

from .. import settings
from ..macro import MacroContext, PassthroughMacroContext


def test_skip_keys(monkeypatch):
    include_key = "_WR_TEST_INCLUDED"
    exclude_key = "_WR_TEST_TOKEN"
    monkeypatch.setattr(settings, "MACRO_INCLUDE_ENV", True)
    try:
        os.environ[exclude_key] = ""
        os.environ[include_key] = ""
        ctx = MacroContext(use_environment=True)
        keys = list(apischema.serialize(MacroContext, ctx)["macros"])
    finally:
        os.environ.pop(include_key, None)
        os.environ.pop(exclude_key, None)

    assert exclude_key not in keys
    assert include_key in keys


def test_skip_long_values(monkeypatch):
    include_key = "_WR_TEST_INCLUDED"
    exclude_key = "_WR_TEST_EXCLUDED"
    monkeypatch.setattr(settings, "MACRO_INCLUDE_ENV", True)
    try:
        os.environ[exclude_key] = "a" * (settings.MACRO_VALUE_MAX_LENGTH + 1)
        os.environ[include_key] = "a" * (settings.MACRO_VALUE_MAX_LENGTH - 2)
        ctx = MacroContext(use_environment=True)
        keys = list(apischema.serialize(MacroContext, ctx)["macros"])
    finally:
        os.environ.pop(include_key, None)
        os.environ.pop(exclude_key, None)

    assert exclude_key not in keys
    assert include_key in keys


@pytest.mark.parametrize("include", [False, True])
def test_env_include_exclude(monkeypatch, include: bool):
    monkeypatch.setattr(settings, "MACRO_INCLUDE_ENV", include)
    ctx = MacroContext(use_environment=True)
    keys = set(apischema.serialize(MacroContext, ctx)["macros"])
    assert not apischema.serialize(MacroContext, ctx)["passthrough"]

    if not include:
        assert set(keys) == set()
    else:
        assert 0 < len(set(keys)) <= len(os.environ)


def test_passthrough():
    ctx = PassthroughMacroContext()
    assert ctx.expand("$(ABC)") == "$(ABC)"

    serialized = apischema.serialize(MacroContext, ctx)
    assert serialized["passthrough"]

    deserialized = apischema.deserialize(MacroContext, serialized)
    assert isinstance(deserialized, PassthroughMacroContext)
    assert deserialized.expand("$(ABC)") == "$(ABC)"



import pathlib

import blark.dependency_store
import pytest

from ..common import IocMetadata
from ..plugins import twincat_pytmc
from . import conftest

BLARK_ROOT = conftest.MODULE_PATH / "blark_root"

if not (BLARK_ROOT / "README.md").exists():
    pytest.fail(
        f"{BLARK_ROOT} is missing: whatrecord was likely not cloned with "
        "--recursive for this test suite submodule"
    )

twincat_projects = list(BLARK_ROOT.glob("**/*.tsproj"))
twincat_ioc_path = conftest.MODULE_PATH / "iocs" / "ioc_twincat"
project_a_path = BLARK_ROOT / "project_a" / "project_a" / "project_a.tsproj"
project_c_path = BLARK_ROOT / "project_c" / "project_c" / "project_c.tsproj"


@pytest.fixture
def project_c_md() -> IocMetadata:
    return IocMetadata(name="ioc_twincat", script=twincat_ioc_path / "st.cmd")


@pytest.fixture
def depstore(
    monkeypatch: pytest.MonkeyPatch,
) -> blark.dependency_store.DependencyStore:
    depstore = blark.dependency_store.DependencyStore(root=BLARK_ROOT)

    def _get_dep_store():
        return depstore

    monkeypatch.setattr(blark.dependency_store, "get_dependency_store", _get_dep_store)
    return depstore


def test_get_project_from_ioc(project_c_md: IocMetadata):
    _, makefile_contents, _ = twincat_pytmc.get_ioc_makefile(project_c_md)
    info = twincat_pytmc.get_project_from_ioc(project_c_md, makefile_contents)
    assert info is not None

    path, plc_name = info
    assert plc_name == "ProjectC"
    assert path == project_c_path


def test_plc_metadata_with_tmc(
    project_c_md: IocMetadata,
    depstore: blark.dependency_store.DependencyStore,
):
    # logging.getLogger("blark").setLevel("DEBUG")
    plc_md, = list(twincat_pytmc.PlcMetadata.from_ioc(project_c_md))

    assert plc_md.context[0].name == str(plc_md.filename)
    assert plc_md.name == "ProjectC"
    assert plc_md.include_dependencies
    expected_files = {
        'MAIN.TcPOU',
        'ProjectC.plcproj',
        'project_c.tsproj',
    }
    loaded_names = {pathlib.Path(file).name for file in plc_md.loaded_files}
    for fn in expected_files:
        assert fn in loaded_names

    assert plc_md.record_to_symbol == {
        # Simple pragma'd LREAL
        "PREFIX:fOutput": "ProjectC:MAIN.fOutput",
        "PREFIX:fOutput_RBV": "ProjectC:MAIN.fOutput",
        # Function block instance which refers back to SampleLibraryA
        "PREFIX:fbTest:fInput": "ProjectC:MAIN.fbTest.fInput",
        "PREFIX:fbTest:fInput_RBV": "ProjectC:MAIN.fbTest.fInput",
        "PREFIX:fbTest:fOutput_RBV": "ProjectC:MAIN.fbTest.fOutput",
    }

    assert plc_md.nc is not None
    assert len(plc_md.nc.axes) == 1
    axis = plc_md.nc.axes[0]
    assert axis.name == "Axis 1"
    assert pathlib.Path(axis.context[0].name).name == "Axis 1.xti"
    assert axis.units == "mm"

    assert set(plc_md.dependencies) == {"SampleLibraryA"}
    assert plc_md.dependencies["SampleLibraryA"] == blark.solution.DependencyVersion(
        name="SampleLibraryA",
        version="*",
        vendor="blark-testing",
        namespace="SampleLibraryA",
    )

    def check_context(md: twincat_pytmc.PlcSymbolMetadata, expected_files: list[str]):
        files = [pathlib.Path(ctx.name).name for ctx in md.context]
        assert set(files) == set(expected_files)

    # Simple symbol, with I/O records:
    sym = plc_md.symbols["ProjectC:MAIN.fOutput"]
    check_context(sym, ["MAIN.TcPOU"])
    assert sym.name == "ProjectC:MAIN.fOutput"
    assert sym.type == "LREAL"
    assert sym.records == ["PREFIX:fOutput_RBV", "PREFIX:fOutput"]

    # Symbol generated from a library FB input, with setpoint/rbv records:
    sym = plc_md.symbols['ProjectC:MAIN.fbTest.fInput']
    check_context(sym, ["MAIN.TcPOU", "FB_SampleA_Test.TcPOU"])
    assert sym.name == "ProjectC:MAIN.fbTest.fInput"
    assert sym.type == "LREAL (SampleLibraryA.FB_SampleA_Test)"
    assert sym.records == ["PREFIX:fbTest:fInput_RBV", "PREFIX:fbTest:fInput"]

    # Symbol generated from a library FB output, with only readback record:
    sym = plc_md.symbols['ProjectC:MAIN.fbTest.fOutput']
    check_context(sym, ["MAIN.TcPOU", "FB_SampleA_Test.TcPOU"])
    assert sym.name == "ProjectC:MAIN.fbTest.fOutput"
    assert sym.type == "LREAL (SampleLibraryA.FB_SampleA_Test)"
    assert sym.records == ["PREFIX:fbTest:fOutput_RBV"]

    assert depstore.get_dependency("project_c") is not None
    assert depstore.get_dependency("SampleLibraryA") is not None


def test_plc_metadata_without_tmc():
    (project_a,) = list(
        twincat_pytmc.PlcMetadata.from_project_filename(
            project_a_path,
            include_dependencies=False,
        )
    )
    assert not project_a.record_to_symbol
    assert not project_a.symbols
    assert "ProjectA.plcproj" in project_a.context[0].name


def test_empty_plugin_results():
    results = twincat_pytmc.TwincatPluginResults.from_metadata_items([])
    assert not results.nested


def test_plugin_results(project_c_md: IocMetadata):
    plc_md, = list(twincat_pytmc.PlcMetadata.from_ioc(project_c_md))
    results = twincat_pytmc.TwincatPluginResults.from_metadata_items(
        # Loading it twice to smoke test the 'merge' function
        [plc_md, plc_md]
    )
    assert results is not None

    # The TwinCAT Plugin uses "nested" metadata, allowing the server to
    # only give you per-project information if you want.  So not everything is
    # in the top-level dictionary.
    assert results.nested is not None
    assert "PREFIX:fOutput_RBV" in results.nested["ProjectC"].record_to_metadata_keys
    assert "ProjectC:MAIN.fOutput" in results.nested["ProjectC"].metadata_by_key



"""
"whatrecord graph" is used to parse and generate a relationship graph of
whatrecord-supported file types.
"""

from __future__ import annotations

import argparse
import logging
import pathlib
import re
import shutil
import tempfile
from typing import List, Optional, Union, cast

import graphviz as gv

from ..common import AnyPath
from ..db import Database
from ..graph import RecordLinkGraph, build_database_relations, graph_links
from ..makefile import DependencyGroup, DependencyGroupGraph, Makefile
from ..shell import LoadedIoc
from ..snl import SequencerProgram
from .parse import parse_from_cli_args

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        "filenames",
        type=str,
        nargs="+",
        help="Script or database filename(s)"
    )

    parser.add_argument(
        "-if",
        "--input-format",
        type=str,
        required=False,
        help=(
            "The file format.  For files that lack a recognized extension or "
            "are otherwise misidentified by whatrecord."
        ),
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        help="Output file to write to.  Defaults to standard output.",
    )

    parser.add_argument(
        "--v3",
        action="store_true",
        help="Use V3 database grammar instead of V4 where possible"
    )

    parser.add_argument(
        "--dbd",
        type=str,
        help="The dbd file, if parsing a database",
    )

    parser.add_argument(
        "--standin-directory",
        type=str,
        nargs="*",
        help='Map a "stand-in" directory to another on disk',
    )

    parser.add_argument(
        "--macros",
        type=str,
        help="Macro to add, in the usual form ``macro=value,...``",
    )

    parser.add_argument(
        '--use-gdb',
        action="store_true",
        help="Use metadata derived from the script binary",
    )

    parser.add_argument(
        "--expand",
        action="store_true",
        help="Expand a substitutions file, as in the msi tool",
    )

    parser.add_argument(
        "--highlight",
        nargs="*",
        type=str,
        default=(".*", ),
        help="Highlight the provided regular expression matches",
    )

    parser.add_argument(
        "--code",
        action="store_true",
        help="Include code information in graph, where applicable"
    )

    return parser


def render_graph_to_file(
    graph: gv.Digraph, default_format: str = "png", filename: Optional[str] = None
):
    if filename is None:
        print(graph.source)
        return

    output_extension = pathlib.Path(filename).suffix
    format = output_extension.lstrip(".").lower() or default_format
    if format == "dot":
        with open(filename, "wt") as fp:
            print(graph.source, file=fp)
    else:
        with tempfile.NamedTemporaryFile(suffix=f".{format}") as source_file:
            rendered_filename = graph.render(
                source_file.name, format=format
            )
        shutil.copyfile(rendered_filename, filename)


DatabaseItem = Union[LoadedIoc, Database]


def _records_by_patterns(
    database: Database,
    patterns: List[str],
) -> List[str]:
    return [
        rec.name
        for rec in database.records.values()
        if any(
            re.search(pattern, rec.name)
            for pattern in patterns
        )
    ]


def get_database_graph(
    *loaded_items: DatabaseItem,
    highlight: Optional[List[str]] = None,
) -> RecordLinkGraph:
    """
    Get a database graph from a number of database items.
    """
    database = Database.from_multiple(*loaded_items)
    starting_records = _records_by_patterns(database, highlight) if highlight else []
    relations = build_database_relations(
        database=database.records,
        record_types=database.record_types or {},
        aliases=database.aliases or {},
    )
    return graph_links(
        database=database.records,
        starting_records=starting_records,
        relations=relations,
    )


def main(
    filenames: List[AnyPath],
    dbd: Optional[str] = None,
    standin_directory: Optional[List[str]] = None,
    macros: Optional[str] = None,
    use_gdb: bool = False,
    input_format: Optional[str] = None,
    expand: bool = False,
    highlight: Optional[List[str]] = None,
    v3: bool = False,
    output: Optional[str] = None,
    code: bool = False,
):
    highlight = highlight or []

    loaded_items = [
        parse_from_cli_args(
            filename=filename,
            dbd=dbd,
            standin_directory=standin_directory,
            macros=macros,
            use_gdb=use_gdb,
            format=input_format,
            expand=expand,
            v3=v3,
        )
        for filename in filenames
    ]

    if all(isinstance(item, (LoadedIoc, Database)) for item in loaded_items):
        loaded_items = cast(List[DatabaseItem], loaded_items)
        graph = get_database_graph(*loaded_items, highlight=highlight)
        if not graph.nodes:
            logger.warning(
                "No records found matching the highlight settings: %s",
                highlight
            )
    else:
        try:
            item, = loaded_items
        except ValueError:
            raise RuntimeError(
                "Sorry, multiple database files are supported but only "
                "single files for other formats."
            )

        if isinstance(item, SequencerProgram):
            graph = item.as_graph(include_code=code)
        elif isinstance(item, Makefile):
            graph = DependencyGroupGraph(DependencyGroup.from_makefile(item))
        else:
            raise RuntimeError(
                f"Sorry, graph isn't supported yet for {item.__class__.__name__}"
            )

    gv_graph = graph.to_digraph()
    logger.debug("Nodes: %s", graph.nodes)
    logger.debug("Edges: %s", graph.edges)
    render_graph_to_file(gv_graph, filename=output)



import asyncio
import threading


class TaskHandler:
    """Helper to manage asyncio tasks in one spot."""

    def __init__(self):
        self.tasks = []
        self._lock = threading.Lock()

    def create(self, coro):
        """Schedule the execution of a coroutine object in a spawn task."""
        task = asyncio.create_task(coro)
        with self._lock:
            self.tasks.append(task)
        task.add_done_callback(self._remove_completed_task)
        return task

    def _remove_completed_task(self, task):
        try:
            with self._lock:
                self.tasks.remove(task)
        except ValueError:
            # May have been cancelled or removed otherwise
            ...

    @staticmethod
    async def cancel(task):
        task.cancel()
        await task

    async def cancel_all(self, wait=False):
        with self._lock:
            tasks = list(self.tasks)
            self.tasks.clear()

        for task in list(tasks):
            task.cancel()

        if wait and len(tasks):
            await asyncio.wait(tasks)



import dataclasses
import pathlib
import subprocess
import textwrap
from typing import Optional, Set, Tuple

import pytest

from .. import makefile
from .conftest import MODULE_PATH, skip_without_make

DEPS_MAKEFILE_ROOT = MODULE_PATH / "deps"

# More coverage on debug log lines
makefile.logger.setLevel("DEBUG")


def prune_result(
    result: makefile.Makefile,
    expected: makefile.Makefile,
    to_keep: Optional[Set[str]] = None,
) -> None:
    """
    Prune resulting Makefile for the purposes of testing.

    * Remove extra environment variables that we don't want here
    """
    for key in set(result.env) - set(expected.env):
        result.env.pop(key)

    ignore = {"env", "filename"}
    for field in dataclasses.fields(makefile.Makefile):
        if field.name not in to_keep and field.name not in ignore:
            setattr(result, field.name, getattr(expected, field.name))


def get_makefile(contents: str, *, set_filename: bool = True) -> makefile.Makefile:
    """Get a Makefile instance given its contents."""
    contents = textwrap.dedent(contents).replace("    ", "\t")
    print("Creating Makefile from contents:")
    print(contents)
    make = makefile.Makefile.from_string(
        contents,
        filename=DEPS_MAKEFILE_ROOT / "Makefile.made_up" if set_filename else None,
    )
    if set_filename:
        # Make the object more realistic
        make.name = "ioc"
        make.variable_name = "ioc_var"
    return make


def get_dependency_group(
    contents: str, *, set_filename: bool = True
) -> Tuple[makefile.Makefile, makefile.DependencyGroup]:
    """Get a DependencyGroup instance given a single Makefile's contents."""
    root = get_makefile(contents, set_filename=set_filename)
    group = makefile.DependencyGroup.from_makefile(root)
    if set_filename:
        assert root.filename is not None
        assert group.root == root.filename.parent
    return root, group


def check_module_in_group(
    group: makefile.DependencyGroup, variable_name: str, path: pathlib.Path
) -> makefile.Dependency:
    """Check if a module is present in a DependencyGroup."""
    if path not in group.all_modules:
        paths = ", ".join(str(path) for path in group.all_modules)
        raise ValueError(
            f"Module {variable_name} not found in dependency list. "
            f"Paths: {paths}"
        )
    module = group.all_modules[path]
    assert module.variable_name == variable_name
    return module


@skip_without_make
def test_make_version():
    version = subprocess.check_output(["make", "--version"])
    print(version)


@skip_without_make
@pytest.mark.parametrize(
    "contents, to_keep, expected",
    [
        pytest.param(
            """
            WHATREC_A=B
            WHATREC_C?=D
            """,
            set(),
            makefile.Makefile(
                env={"WHATREC_A": "B", "WHATREC_C": "D"},
                filename=None,
            ),
            id="simple-subst-0",
        ),
        pytest.param(
            """
            WHATREC_A=A
            WHATREC_B=$(WHATREC_A)
            WHATREC_C?=C
            """,
            set(),
            makefile.Makefile(
                env={"WHATREC_A": "A", "WHATREC_B": "A", "WHATREC_C": "C"},
                filename=None,
            ),
            id="simple-subst-1",
        ),
        pytest.param(
            """
            ENV_VAR=A

            all:
                echo "Hi!"
            """,
            {"default_goal", },
            makefile.Makefile(
                env={"ENV_VAR": "A"},
                default_goal="all",
                filename=None,
            ),
            id="default-goal",
        ),
        pytest.param(
            # These would typically be set by the epics build system; but let's
            # set them ourselves as this is just a test of our stuff and not
            # the EPICS build system:
            """
            BUILD_ARCHS=arch1 arch2
            CROSS_COMPILER_HOST_ARCHS=arch3 arch4
            CROSS_COMPILER_TARGET_ARCHS=arch5 arch6
            BASE_MODULE_VERSION=R7.0.2
            CONFIG=/
            """,
            {
                "build_archs",
                "cross_compiler_host_archs",
                "cross_compiler_target_archs",
                "base_version",
                "base_config_path",
            },
            makefile.Makefile(
                env={},
                build_archs=["arch1", "arch2"],
                cross_compiler_host_archs=["arch3", "arch4"],
                cross_compiler_target_archs=["arch5", "arch6"],
                base_version="R7.0.2",
                base_config_path=pathlib.Path("/"),
                filename=None,
            ),
            id="epics-build-system-vars",
        ),
    ]
)
def test_from_contents(contents: str, expected: makefile.Makefile, to_keep: Set[str]):
    result = get_makefile(contents, set_filename=False)
    print(result)
    prune_result(result, expected, to_keep=to_keep)
    assert result == expected


@skip_without_make
@pytest.mark.parametrize(
    "contents",
    [
        pytest.param(
            f"""
            EPICS_BASE={DEPS_MAKEFILE_ROOT}/base
            RELEASE_TOP=EPICS_BASE
            """,
            id="initial",
        ),
    ]
)
def test_dependency_norecurse_smoke(contents: str):
    make = get_makefile(contents)
    dep = makefile.Dependency.from_makefile(make, name="name", variable_name="var")
    assert dep.makefile == make
    assert dep.name == "name"
    assert dep.variable_name == "var"


@skip_without_make
def test_dependency_group_base_only():
    _, group = get_dependency_group(
        f"""
        EPICS_BASE={DEPS_MAKEFILE_ROOT}/base
        RELEASE_TOPS=EPICS_BASE
        """
    )

    assert len(group.all_modules) == 2

    base = check_module_in_group(group, "EPICS_BASE", DEPS_MAKEFILE_ROOT / "base")
    # I'm not convinced this is the right behavior, but not sure if we can
    # do better generically:
    assert base.name == "EPICS_BASE"


@skip_without_make
def test_dependency_group_base_and_module_a():
    _, group = get_dependency_group(
        f"""
        EPICS_BASE={DEPS_MAKEFILE_ROOT}/base
        MODULE_A={DEPS_MAKEFILE_ROOT}/module_a
        RELEASE_TOPS=EPICS_BASE MODULE_A
        """
    )

    assert len(group.all_modules) == 3
    check_module_in_group(group, "EPICS_BASE", DEPS_MAKEFILE_ROOT / "base")
    check_module_in_group(group, "MODULE_A", DEPS_MAKEFILE_ROOT / "module_a")


@skip_without_make
def test_dependency_group_base_and_layers():
    _, group = get_dependency_group(
        f"""
        EPICS_BASE={DEPS_MAKEFILE_ROOT}/base
        MODULE_C={DEPS_MAKEFILE_ROOT}/module_c
        RELEASE_TOPS=EPICS_BASE MODULE_C
        """
    )

    # test suite -> module_c   -> module_a -> base
    #         |                -> module_b -> base
    #         -> base
    #

    base = check_module_in_group(group, "EPICS_BASE", DEPS_MAKEFILE_ROOT / "base")
    ma = check_module_in_group(group, "MODULE_A", DEPS_MAKEFILE_ROOT / "module_a")
    mb = check_module_in_group(group, "MODULE_B", DEPS_MAKEFILE_ROOT / "module_b")
    mc = check_module_in_group(group, "MODULE_C", DEPS_MAKEFILE_ROOT / "module_c")
    ioc_path = DEPS_MAKEFILE_ROOT  # <-- test suite pseudo-makefile
    assert set(group.all_modules) == {
        ioc_path,
        base.path,
        ma.path,
        mb.path,
        mc.path,
    }
    assert len(group.all_modules) == 5

    # Base
    assert base.dependents == {
        "MODULE_A": ma.path,
        "MODULE_B": mb.path,
        "MODULE_C": mc.path,
        None: ioc_path,
    }
    assert len(base.dependencies) == 0

    # module a
    assert ma.dependencies == {"EPICS_BASE": base.path}
    assert ma.dependents == {"MODULE_C": mc.path}

    # module b
    assert mb.dependencies == {"EPICS_BASE": base.path}
    assert mb.dependents == {"MODULE_C": mc.path}

    # module c
    assert mc.dependencies == {
        "EPICS_BASE": base.path,
        "MODULE_A": ma.path,
        "MODULE_B": mb.path,
    }
    assert mc.dependents == {None: ioc_path}


@skip_without_make
def test_dependency_group_graph():
    _, group = get_dependency_group(
        f"""
        EPICS_BASE={DEPS_MAKEFILE_ROOT}/base
        MODULE_C={DEPS_MAKEFILE_ROOT}/module_c
        RELEASE_TOPS=EPICS_BASE MODULE_C
        """
    )
    graph = group.as_graph()
    assert len(graph.nodes) == 5

    base = check_module_in_group(group, "EPICS_BASE", DEPS_MAKEFILE_ROOT / "base")
    ma = check_module_in_group(group, "MODULE_A", DEPS_MAKEFILE_ROOT / "module_a")
    mb = check_module_in_group(group, "MODULE_B", DEPS_MAKEFILE_ROOT / "module_b")
    mc = check_module_in_group(group, "MODULE_C", DEPS_MAKEFILE_ROOT / "module_c")

    edge_pairs = list(graph.edge_pairs)

    def check_edge(source: makefile.Dependency, dest: makefile.Dependency):
        node_source = graph.get_node(str(source.path))
        node_dest = graph.get_node(str(dest.path))
        assert (node_source, node_dest) in edge_pairs

    check_edge(source=ma, dest=base)
    check_edge(source=mb, dest=base)
    check_edge(source=mc, dest=base)
    check_edge(source=mc, dest=ma)
    check_edge(source=mc, dest=mb)

    # Smoke test to_digraph
    graph.to_digraph()



"""
"whatrecord server" is used to start an aiohttp-backed web server which hosts
startup script and record information.
"""

import argparse
import logging

from ..server import main

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        '--scripts',
        type=str,
        nargs='*',
        help="Startup script filename(s)"
    )

    parser.add_argument(
        '--script-loader',
        type=str,
        nargs='*',
        help="Run an external script to get IOC configuration information",
    )

    archive_group = parser.add_mutually_exclusive_group()

    archive_group.add_argument(
        '--archive-management-url',
        type=str,
        help='Archiver management URL for finding archived PVs.'
    )

    parser.add_argument(
        '--gateway-config',
        type=str,
        help='Gateway configuration file or directory'
    )

    parser.add_argument(
        '--standin-directory',
        type=str,
        nargs='*',
        help='Map a "stand-in" directory to another on disk'
    )

    parser.add_argument(
        '--port',
        type=int,
        default=8898,
        help='Web server TCP port'
    )

    parser.add_argument(
        '--tracemalloc',
        dest='use_tracemalloc',
        action='store_true',
        help='[Debug] Use tracemalloc to debug server memory usage'
    )

    parser.add_argument(
        '--dump-for-offline-usage',
        dest="offline_dump_target",
        type=str,
        help='Dump the server state for offline usage of the whatrecord frontend'
    )

    parser.add_argument(
        "--partial-dump",
        action="store_true",
        help="Reduce size of the dump by removing less important things (.dbd files, etc)"
    )

    return parser


__all__ = ["build_arg_parser", "main"]



import pytest

from .. import util
from ..common import LoadContext
from ..gateway import (AccessSecurity, AliasRule, AllowRule, DenyRule,
                       GatewayConfig, PVList, PVListMatch, PVListMatches)
from .conftest import MODULE_PATH


@pytest.fixture
def kfe_pvlist():
    return PVList.from_file(
        (MODULE_PATH / "kfe.pvlist").resolve()
    )


@pytest.fixture
def gateway_config():
    return GatewayConfig(path=MODULE_PATH, glob_str="*.pvlist")


@pytest.mark.parametrize(
    "rule, expected",
    [
        pytest.param(
            "SXR:YAG:CVV:01.* ALLOW RWINSTRMCC 1",
            AllowRule(
                context=(LoadContext("None", 1), ),
                command="ALLOW",
                pattern="SXR:YAG:CVV:01.*",
                access=AccessSecurity(
                    group="RWINSTRMCC",
                    level="1",
                )
            ),
            id="allow_basic",
        ),
        pytest.param(
            "SXR:YAG:CVV:01.* DENY",
            DenyRule(
                context=(LoadContext("None", 1), ),
                command="DENY",
                pattern="SXR:YAG:CVV:01.*",
                hosts=[],
            ),
            id="deny_basic",
        ),
        pytest.param(
            "SXR:YAG:CVV:01.* DENY FROM host1",
            DenyRule(
                context=(LoadContext("None", 1), ),
                command="DENY",
                pattern="SXR:YAG:CVV:01.*",
                hosts=["host1"],
            ),
            id="deny_host",
        ),
        pytest.param(
            "SXR:YAG:CVV:01.* DENY FROM host1 host2",
            DenyRule(
                context=(LoadContext("None", 1), ),
                command="DENY",
                pattern="SXR:YAG:CVV:01.*",
                hosts=["host1", "host2"],
            ),
            id="deny_hosts",
        ),
        pytest.param(
            r":gateway\.\(.*\)Flag        ALIAS     gateway:\1Flag       ",
            AliasRule(
                context=(LoadContext("None", 1), ),
                command="ALIAS",
                pattern=r":gateway\.\(.*\)Flag",
                pvname=r"gateway:\1Flag",
                access=None,
            ),
            id="alias",
        ),
        pytest.param(
            r":gateway\.\(.*\)Flag        ALIAS     gateway:\1Flag          GatewayAdmin",
            AliasRule(
                context=(LoadContext("None", 1), ),
                command="ALIAS",
                pattern=r":gateway\.\(.*\)Flag",
                pvname=r"gateway:\1Flag",
                access=AccessSecurity(
                    group="GatewayAdmin",
                    level=None,
                )
            ),
            id="alias_access",
        ),

        pytest.param(
            "*        ALIAS     gateway:Flag",
            AliasRule(
                context=(LoadContext("None", 1), ),
                command="ALIAS",
                pattern="*",
                pvname="gateway:Flag",
                # Hard to test for, as this is set automatically:
                # metadata={
                #     "error": (
                #         "Invalid regex. error: nothing to "
                #         "repeat at position 0'",
                #     )
                # }
            ),
            id="invalid_pattern",
        ),

    ]
)
def test_rule(rule, expected):
    assert PVList.from_string(rule).rules[0] == expected


def test_eval_order():
    assert PVList.from_string(
        "evaluation order allow, deny"
    ).evaluation_order == "ALLOW, DENY"


def test_full():
    source = """
        # Header Text 1
        # Header Text 2
        evaluation order allow, deny

        # Skipped Comment

        # Rule Comment 1
        # Rule Comment 2
        SXR:YAG:CVV:01.* ALLOW RWINSTRMCC 1
    """

    source_hash = util.get_bytes_sha256(source.encode("utf-8"))
    assert PVList.from_string(
        source,
        filename="filename",
    ) == PVList(
        filename="filename",
        evaluation_order="ALLOW, DENY",
        comments=[
            "Header Text 1",
            "Header Text 2",
            "Skipped Comment",
            "Rule Comment 1",
            "Rule Comment 2",
        ],
        header="Header Text 1\nHeader Text 2",
        hash=source_hash,
        rules=[
            AllowRule(
                context=(LoadContext("filename", 9), ),
                command="ALLOW",
                pattern="SXR:YAG:CVV:01.*",
                access=AccessSecurity(
                    group="RWINSTRMCC",
                    level="1",
                ),
                header="Rule Comment 1\nRule Comment 2",
            ),
        ],
    )


def test_match_without_context(kfe_pvlist: PVList):
    assert list(kfe_pvlist.match("IOC:KFE:ABC")) == [
        (
            AllowRule(
                context=(LoadContext(kfe_pvlist.filename, 10,), ),
                pattern="[A-Z][A-Z][A-Z]:KFE:.*",
                command="ALLOW",
                # regex=re.compile("[A-Z][A-Z][A-Z]:KFE:.*"),
                header="",
                metadata={},
                access=None,
            ),
            []
        )
    ]


def test_match_with_context(kfe_pvlist: PVList):
    assert list(kfe_pvlist.match("IOC:RIX:ABC")) == [
        (
            AllowRule(
                context=(LoadContext(kfe_pvlist.filename, 13,), ),
                pattern="[A-Z][A-Z][A-Z]:RIX:.*",
                command="ALLOW",
                # regex=re.compile("[A-Z][A-Z][A-Z]:KFE:.*"),
                header="Some RIX devices temporarily on KFE subnet",
                metadata={},
                access=None,
            ),
            []
        )
    ]


def test_gateway_config(kfe_pvlist: PVList, gateway_config: GatewayConfig):
    gateway_config.update_changed()

    kfe_fn = str(kfe_pvlist.filename)
    assert gateway_config.get_matches("IOC:RIX:ABC") == PVListMatches(
        name="IOC:RIX:ABC",
        matches=[
            PVListMatch(
                filename=kfe_fn,
                rule=AllowRule(
                    context=(LoadContext(kfe_fn, 13,), ),
                    pattern="[A-Z][A-Z][A-Z]:RIX:.*",
                    command="ALLOW",
                    # regex=re.compile("[A-Z][A-Z][A-Z]:KFE:.*"),
                    header="Some RIX devices temporarily on KFE subnet",
                    metadata={},
                    access=None,
                ),
                groups=[],
            ),
        ]
    )



"""
TwinCAT / pytmc / blark whatrecord plugin

Match your TwinCAT project symbols to EPICS records.

Set ``project`` to generate metadata for one particular PLC project.

Alternatively, the plugin will query WHATRECORD_SERVER for all IOCs and attempt
to find ads-ioc-based IOCs with associated TwinCAT projects.

To annotate library-level information in dependencies (i.e., to let users "go
to the source code of dependency-library-sourced symbols"), be sure to
set ``BLARK_TWINCAT_ROOT`` per blark's documentation.
"""
from __future__ import annotations

import argparse
import asyncio
import dataclasses
import json
import logging
import pathlib
import re
import typing
from dataclasses import dataclass
from typing import Dict, Generator, Iterable, List, Optional, Tuple, cast

import apischema
import blark
import blark.dependency_store
import blark.summary
import pytmc.parser

from .. import cache, client, util
from ..common import AnyPath, FullLoadContext, IocMetadata, LoadContext
from ..server.common import PluginResults
from ..util import get_file_sha256, read_text_file_with_hash

logger = logging.getLogger(__name__)

# Stash the description for later usage by the CLI interface.
DESCRIPTION = __doc__.strip()


def _stringify_path(path: AnyPath) -> str:
    """Uniform strings from a pathlib.Path or string."""
    if isinstance(path, str):
        path = pathlib.Path(path)
    return str(path.resolve())


@dataclass
class TwincatPluginResults(PluginResults):
    def merge(self, results: TwincatPluginResults) -> None:
        self.files_to_monitor.update(results.files_to_monitor)
        self.record_to_metadata_keys.update(results.record_to_metadata_keys)
        self.metadata_by_key.update(results.metadata_by_key)
        self.execution_info.update(results.execution_info)
        if self.nested is not None:
            self.nested.update(results.nested or {})

    @classmethod
    def from_metadata(cls, md: PlcMetadata) -> TwincatPluginResults:
        plc_results = TwincatPluginResults(
            files_to_monitor={
                _stringify_path(path): shasum
                for path, shasum in md.loaded_files.items()
            },
            record_to_metadata_keys={
                rec: [sym] for rec, sym in md.record_to_symbol.items()
            },
            metadata_by_key=md.symbols,
            metadata={
                "dependencies": md.dependencies,
                "nc": md.nc,
            },
            execution_info={},
        )
        return TwincatPluginResults(
            nested={
                md.name: plc_results
            }
        )

    @classmethod
    def from_metadata_items(
        cls, mds: Iterable[PlcMetadata]
    ) -> TwincatPluginResults:
        results = None
        for plc_md in mds:
            single = TwincatPluginResults.from_metadata(plc_md)
            if results is None:
                results = single
            else:
                results.merge(single)

        if results is not None:
            return results

        return TwincatPluginResults(
            files_to_monitor={},
            record_to_metadata_keys={},
            metadata_by_key={},
            execution_info={"result": "No PLCs found."},
        )


@dataclass
class Declaration:
    context: FullLoadContext
    name: str = "unknown"
    type: str = "unknown"


@dataclass
class DataType:
    name: str


@dataclass
class PlcCode:
    ...


@dataclass
class NCAxis:
    """A single NC axis."""
    context: FullLoadContext
    filename: str
    hash: str

    axis_id: int
    name: str
    units: str
    params: Dict[str, str]

    @classmethod
    def from_pytmc(
        cls,
        axis: pytmc.parser.Axis,
    ) -> NCAxis:
        filename = str(axis.filename.resolve())
        return cls(
            context=(LoadContext(filename, 0), ),
            filename=filename,
            hash=get_file_sha256(filename),
            axis_id=axis.axis_number,
            name=axis.name,
            units=axis.units,
            params=dict(axis.summarize()),
        )


@dataclass
class NCAxes:
    """Top-level NC axis information."""
    context: FullLoadContext
    filename: pathlib.Path
    hash: str
    axes: List[NCAxis]

    @classmethod
    def from_pytmc(
        cls,
        project: pytmc.parser.TwincatItem,
        loaded_files: Optional[Dict[str, str]] = None,
    ) -> Optional[NCAxes]:
        """Create an NCAxes instance from a project filename."""
        try:
            nc: pytmc.parser.NC = next(project.root.find(pytmc.parser.NC))
        except StopIteration:
            return None

        nc_filename = _stringify_path(nc.filename)
        nc = cls(
            context=(LoadContext(nc_filename, 0), ),
            filename=nc.filename,
            hash=get_file_sha256(nc_filename),
            axes=[NCAxis.from_pytmc(axis) for axis in nc.axes],
        )

        if loaded_files is not None:
            loaded_files[_stringify_path(nc.filename)] = nc.hash
            for axis in nc.axes:
                loaded_files[_stringify_path(axis.filename)] = axis.hash

        return nc

    @classmethod
    def from_project_filename(
        cls,
        filename: pathlib.Path,
        loaded_files: Optional[Dict[str, str]] = None,
    ) -> Optional[NCAxes]:
        """Create an NCAxes instance from a project filename."""
        project = pytmc.parser.parse(filename)
        return cls.from_pytmc(project, loaded_files)


@dataclass
class PlcSymbolMetadata:
    context: FullLoadContext
    name: str
    type: str
    records: List[str]


@dataclass
class PlcMetadataCacheKey(cache.CacheKey):
    """
    These attributes define a PlcMetadata cache item.

    The PLC name and filename will be used as a cache key; however, additional
    checks will be made to see that the files have not changed on disk since
    the last save.
    """
    name: str
    filename: str
    include_dependencies: bool


def load_context_from_path(path: List[blark.summary.Summary]) -> FullLoadContext:
    """Get a FullLoadContext from a blark variable path."""
    result = []
    saw_files = set()
    for file, line in reversed(blark.summary.path_to_file_and_line(path)):
        file = str(file)
        if file not in saw_files:
            result.append(LoadContext(str(file), line))
            saw_files.add(file)

    return tuple(result[::-1])


def get_symbol_metadata(
    blark_md: blark.dependency_store.PlcMetadata,
    symbol: pytmc.parser.Symbol,
    require_records: bool = True,
    add_project_prefix: bool = True,
) -> Generator[PlcSymbolMetadata, None, None]:
    """Get symbol metadata given a pytmc Symbol."""
    symbol_type_name = symbol.data_type.qualified_type_name
    for pkg in pytmc.pragmas.record_packages_from_symbol(
        symbol,
        yield_exceptions=True,
        allow_no_pragma=False,
    ):
        if isinstance(pkg, Exception):
            # Eat these up rather than raising
            continue
        # context = get_symbol_context(symbol.data_type.name, pkg.tcname)
        path = blark_md.summary.find_path(pkg.tcname)
        if not path:
            # Can't find the declaration in the code, somehow
            continue

        records = [record.pvname for record in pkg.records]
        if records or not require_records:
            annotated_name = pkg.tcname
            if add_project_prefix:
                annotated_name = f"{blark_md.name}:{annotated_name}"

            try:
                chain_type_name = pkg.chain.data_type.qualified_type_name
            except AttributeError:
                chain_type_name = "unknown"

            if symbol_type_name == chain_type_name:
                type_name = symbol_type_name
            else:
                type_name = f"{chain_type_name} ({symbol_type_name})"

            yield PlcSymbolMetadata(
                context=load_context_from_path(path),
                name=annotated_name,
                records=records,
                type=type_name,
            )


@dataclass
class PytmcMetadata:
    tsproj: pytmc.parser.TcSmProject
    plcproj: pytmc.parser.PlcProject
    tmc: Optional[pytmc.parser.TcModuleClass] = None
    nc: Optional[NCAxes] = None

    @classmethod
    def from_tsproj(
        cls,
        tsproj: blark.solution.Project,
        plc_name: str,
        *,
        loaded_files: Optional[Dict[str, str]] = None,
    ) -> PytmcMetadata:
        if loaded_files is None:
            loaded_files = {}
        if tsproj.local_path is None:
            raise ValueError("Project without path")

        try:
            pytmc_tsproj = cast(
                pytmc.parser.TcSmProject,
                pytmc.parser.parse(tsproj.local_path)
            )
        except Exception as ex:
            raise ValueError(f"Failed to parse project: {tsproj.name}") from ex

        try:
            pytmc_plc = cast(
                pytmc.parser.PlcProject,
                pytmc_tsproj.plcs_by_name[plc_name]
            )
        except KeyError as ex:
            raise ValueError(f"Unable to find PLC project: {tsproj.name}") from ex

        md = cls(
            tsproj=pytmc_tsproj,
            plcproj=pytmc_plc,
            tmc=getattr(pytmc_plc, "tmc", None),
        )

        try:
            md.nc = NCAxes.from_pytmc(
                pytmc_tsproj,
                loaded_files=loaded_files,
            )
        except Exception:
            logger.exception(
                "Failed to load NC settings from project: %s",
                tsproj.local_path,
            )
        return md


@dataclass
class PlcMetadata(cache.InlineCached, PlcMetadataCacheKey):
    """This metadata is keyed on PlcMetadataCacheKey."""
    context: FullLoadContext
    symbols: Dict[str, PlcSymbolMetadata]
    loaded_files: Dict[str, str]
    record_to_symbol: Dict[str, str]
    # dependencies: Dict[str, blark.solution.DependencyInformation]
    # TODO
    dependencies: Dict[str, blark.dependency_store.DependencyVersion]
    blark_md: Optional[blark.dependency_store.PlcProjectMetadata] = dataclasses.field(
        default=None,
        metadata=apischema.metadata.skip
    )
    tsproj: Optional[blark.solution.Project] = dataclasses.field(
        default=None,
        metadata=apischema.metadata.skip
    )
    nc: Optional[NCAxes] = None

    @classmethod
    def from_ioc(
        cls,
        md: IocMetadata,
    ) -> Generator[PlcMetadata, None, None]:
        try:
            makefile_hash, makefile_contents, makefile_path = get_ioc_makefile(md)
        except FileNotFoundError:
            return

        project_info = get_project_from_ioc(md, makefile_contents)
        if project_info is None:
            logger.debug("No project found for %s", md.name)
            return

        loaded_files = {
            str(makefile_path): makefile_hash or get_file_sha256(makefile_path)
        }

        project, plc_name = project_info
        logger.info(
            "Found a PLC for this project: %s %s (%s)",
            md.name, plc_name, project
        )
        yield from PlcMetadata.from_project_filename(
            project,
            plc_whitelist=[plc_name],
            loaded_files=loaded_files,
        )

    @classmethod
    def from_blark(
        cls,
        blark_md: blark.dependency_store.PlcProjectMetadata,
        tsproj: blark.solution.Project,
        include_dependencies: bool = True,
        use_cache: bool = True,
    ) -> PlcMetadata:
        """Create a PlcMetadata instance from a pytmc-parsed one."""
        loaded_files = {
            _stringify_path(path): hash_
            for path, hash_ in blark_md.loaded_files.items()
        }

        pytmc_md = PytmcMetadata.from_tsproj(
            tsproj,
            plc_name=blark_md.name,
            loaded_files=loaded_files,
        )

        # TODO: blark - not capturing tsproj
        if tsproj.local_path is not None:
            loaded_files[_stringify_path(tsproj.local_path)] = get_file_sha256(
                tsproj.local_path
            )

        plc_fn = _stringify_path(blark_md.filename)
        if blark_md.plc is None or pytmc_md.tmc is None:
            logger.debug(
                "%s (%s): No TMC file for symbols; skipping...",
                plc_fn,
                blark_md.name,
            )
            return cls(
                name=blark_md.name,
                filename=plc_fn,
                include_dependencies=include_dependencies,
                context=(LoadContext(plc_fn, 0), ),
                dependencies=blark_md.dependencies,
                symbols={},
                record_to_symbol={},
                nc=None,
                loaded_files=loaded_files,
                tsproj=tsproj,
                blark_md=blark_md,
            )

        if use_cache:
            key = PlcMetadataCacheKey(
                name=blark_md.name,
                filename=plc_fn,
                include_dependencies=include_dependencies,
            )
            cached = cls.from_cache(key)
            if cached is not None:
                if util.check_files_up_to_date(cached.loaded_files):
                    return cached

        loaded_files[plc_fn] = util.get_file_sha256(plc_fn)

        md = cls(
            name=blark_md.plc.name or blark_md.name,
            filename=plc_fn,
            include_dependencies=include_dependencies,
            context=(LoadContext(plc_fn, 0), ),
            symbols={},
            record_to_symbol={},
            dependencies=blark_md.dependencies,
            loaded_files=loaded_files,
            nc=pytmc_md.nc,
            tsproj=tsproj,
            blark_md=blark_md,
        )

        def by_name(symbol):
            return symbol.name

        tmc = pytmc_md.tmc
        for symbol in sorted(pytmc.pragmas.find_pytmc_symbols(tmc), key=by_name):
            for symbol_md in get_symbol_metadata(blark_md, symbol):
                md.symbols[symbol_md.name] = symbol_md
                for record in symbol_md.records:
                    md.record_to_symbol[record] = symbol_md.name

        logger.debug(
            "PLC %s: Found %d symbols which generated metadata -> %d records",
            blark_md.plc.name,
            len(md.symbols),
            len(md.record_to_symbol),
        )
        if use_cache:
            md.save_to_cache()

        return md

    @classmethod
    def from_project_filename(
        cls,
        project: AnyPath,
        include_dependencies: bool = True,
        plc_whitelist: Optional[List[str]] = None,
        loaded_files: Optional[Dict[str, str]] = None,
        use_cache: bool = True,
    ) -> Generator[PlcMetadata, None, None]:
        """Given a project/solution filename, get all PlcMetadata."""
        if loaded_files is None:
            loaded_files = {}
        solution = blark.solution.make_solution_from_files(project)
        logger.debug(
            "Solution path %s projects %s", solution.filename, solution.projects
        )
        for tsproj_project in solution.projects:
            logger.debug("Found tsproj %s", tsproj_project.name)
            try:
                parsed_tsproj = tsproj_project.load()
            except Exception:
                logger.exception("Failed to load project %s", tsproj_project.name)
                continue

            for plc_name, plc in parsed_tsproj.plcs_by_name.items():
                if plc_whitelist and plc_name not in plc_whitelist:
                    continue

                logger.debug("Found PLC project %s", plc_name)
                plc_md = blark.dependency_store.PlcProjectMetadata.from_plcproject(
                    plc,
                    include_dependencies=include_dependencies,
                )
                if plc_md is None:
                    continue

                logger.debug(
                    "tsproj %s: Found plc project %s from %s",
                    tsproj_project.name,
                    plc_md.name,
                    plc_md.filename,
                )
                yield cls.from_blark(
                    blark_md=plc_md,
                    tsproj=tsproj_project,
                    include_dependencies=include_dependencies,
                    use_cache=use_cache,
                )


MAKEFILE_VAR_RE = re.compile(r"^([A-Z_][A-Z0-9_]+)\s*:?=\s*(.*)$", re.IGNORECASE | re.MULTILINE)


def get_ioc_makefile(md: IocMetadata) -> Tuple[str, str, pathlib.Path]:
    """Get the IOC Makefile contents, if available."""
    makefile_path = (md.script.parent / "Makefile").resolve()
    sha, contents = read_text_file_with_hash(makefile_path)
    return sha, contents, makefile_path


def get_project_from_ioc(md: IocMetadata, makefile: str) -> Optional[Tuple[pathlib.Path, str]]:
    """Get the TwinCAT Project from a provided ads-ioc IocMetadata and Makefile contents."""
    variables = dict(MAKEFILE_VAR_RE.findall(makefile))
    logger.debug("IOC: %s Makefile variables: %s", md.name, variables)
    try:
        plc_name = variables["PLC"]
        project_path = variables["PROJECT_PATH"]
    except KeyError:
        return None

    project_path = (md.script.parent / project_path).resolve()
    if not project_path.exists():
        logger.debug("Project path doesn't exist: %s", project_path)
        return

    return project_path, plc_name


async def main(
    project: str,
    server: Optional[str] = None,
    pretty: bool = False,
    verbose: bool = False,
) -> List[PlcMetadata]:
    if verbose:
        logging.basicConfig(level="DEBUG")
        logging.getLogger("parso").setLevel("WARNING")

    if project is not None:
        return list(PlcMetadata.from_project_filename(project))

    results = []
    ioc_info = await client.get_iocs(server=server)
    for md in ioc_info.matches:
        try:
            for plc_md in PlcMetadata.from_ioc(md):
                results.append(plc_md)
        except Exception:
            logger.exception("Failed to add PLC from makefile: %s", md.name)
    return results


async def _cli_main():
    parser = _get_argparser()
    args = parser.parse_args()
    results = await main(**vars(args))
    whatrecord_results = TwincatPluginResults.from_metadata_items(results)
    json_results = apischema.serialize(whatrecord_results)
    dump_args = {"indent": 4} if args.pretty else {}
    print(json.dumps(json_results, sort_keys=True, **dump_args))


def _get_argparser(parser: typing.Optional[argparse.ArgumentParser] = None):
    if parser is None:
        parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        "project",
        nargs="?",
        help="TwinCAT Project to parse"
    )
    parser.add_argument(
        "--server", help="WhatRecord API server URL"
    )
    parser.add_argument(
        "-p", "--pretty", action="store_true", help="Pretty JSON output"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Verbose logging"
    )
    return parser


if __name__ == "__main__":
    results = asyncio.run(_cli_main())



"""PVA/V4-related parsing tests"""

import math
import textwrap
from typing import Optional

import lark
import pytest

from ..common import LoadContext
from ..db import Database, PVAFieldReference, RecordField, RecordInstance
from ..format import FormatContext, FormatOptions


def test_simple():
    db = Database.from_string(
        """
record(ai, "rec:X") {
  info(Q:group, {
    "grp:name": {
        "X": {+channel:"VAL"}
    }
  })
}
record(ai, "rec:Y") {
  info(Q:group, {
    "grp:name": {
        "Y": {+channel:"VAL"}
    }
  })
}
""")
    assert db.pva_groups["grp:name"] == RecordInstance(
        context=(LoadContext("None", 4), ),
        record_type="PVA",
        name="grp:name",
        is_pva=True,
        fields={
            "X": PVAFieldReference(
                context=(LoadContext("None", 5), ),
                name="X",
                record_name="rec:X",
                field_name="VAL",
                metadata={"+channel": "VAL"},
            ),
            "Y": PVAFieldReference(
                context=(LoadContext("None", 12), ),
                name="Y",
                record_name="rec:Y",
                field_name="VAL",
                metadata={"+channel": "VAL"},
            ),
        }
    )


def test_link_equivalence():
    # record(longin, "src") {
    #     field(INP, {pva:{
    #         pv:"tgt",
    #         field:"",
    #         local:false,
    #         Q:4,
    #         pipeline:false,
    #         proc:none,
    #         sevr:false,
    #         time:false,
    #         monorder:0,
    #         retry:false,
    #         always:false,
    #         defer:false
    #     }})
    # }
    Database.from_string(
        """
record(longin, "tgt1") {}
record(longin, "src1") {
    field(INP, {pva:"tgt1"})
}

record(longin, "tgt2") {}
record(longin, "src2") {
    field(INP, {pva:{pv:"tgt2"}})
}
"""
    )

    # TODO check src1 vs src2


def test_other_stuff():
    Database.from_string(
        """
record(ai, "time_tag") {
  info(Q:time:tag, "nsec:lsb:20")
}

record(ai, "display_form_hint") {
  info(Q:form, "Default") # implied default
}

record(longin, "tgt3") {}
record(longin, "src3") {
    field(INP, {pva:{
        pv:"tgt3",
        field:"",   # may be a sub-field
        local:false,# Require local PV
        Q:4,        # monitor queue depth
        pipeline:false, # require that server uses monitor flow control protocol
        proc:none,  # Request record processing (side-effects).
        sevr:false, # Maximize severity.
        time:false, # set record time during getValue
        monorder:0, # Order of record processing as a result of CP and CPP
        retry:false,# allow Put while disconnected.
        always:false,# CP/CPP input link process even when .value field hasn't changed
        defer:false # Defer put
    }})
}
"""
    )


@pytest.mark.xfail
def test_numbers():
    db = Database.from_string(
        """
record(ai, "value") {
    field(NAN, NaN)
    field(HEXINT, 0x10)
    field(VAL, {"a": 1e10})
}
""")

    # I think json_string is taking priority, meaning we faithfully get
    # back the strings, but we don't interpret the "number" rule at all.
    # All numbers then become strings, which isn't exactly ideal, but
    # also not a deal-breaker for the most part.
    assert db.records["value"].fields["NAN"].value is math.nan
    assert db.records["value"].fields["HEXINT"].value == 0x10
    assert db.records["value"].fields["VAL"].value == {"a": 1e10}


def test_tab_in_field():
    # Believe it or not, tabs are not accepted in fields in V4 (V3 is more lax)
    with pytest.raises(lark.UnexpectedToken):
        Database.from_string(
            """\
record(ai, "rec:X") {
    field(A, "test\tvalue")
}
"""
        )


def test_list_as_value():
    db = Database.from_string(
        """\
record(ai, "rec:X") {
    field(VAL, ["test value"])
}
""",
        version=4,
    )
    val = db.records["rec:X"].fields["VAL"]
    assert isinstance(val, RecordField)
    assert val.value == ("test value",)


@pytest.mark.parametrize(
    "input_text, expected",
    [
        pytest.param(
            """\
            menu(stringoutPOST) {
                choice(stringoutPOST_OnChange, "On Change")
                choice(stringoutPOST_Always, "Always")
            }
            recordtype(ai) {
                %#include "epicsTypes.h"
                %#include "link.h"
                field(NAME, DBF_STRING) {
                    prompt("Record Name")
                    special(SPC_NOMOD)
                    size(61)
                }
            }
            device(ai, CONSTANT, devAaiSoft, "Soft Channel")
            link(state, lnkStateIf)
            registrar(rsrvRegistrar)
            variable(dbBptNotMonotonic, int)
            """,
            None,
            id="dbd_basic"
        ),
        pytest.param(
            """\
            record(ai, "rec:X") {
                field(A, "test")
                field(B, "test")
                info(info, "X")
            }
            """,
            None,
            id="record_basic"
        ),
        pytest.param(
            """\
            record(ai, "rec:X") {
                field(VAL, [
                    "test value"
                ])
            }
            """,
            None,
            id="record_array"
        ),
        # TODO: quotes indicate transformer or grammar priority issue?
        pytest.param(
            """\
            record(ai, "value") {
                field(HEXINT, 0x10)
            }
            """,
            None,
            marks=pytest.mark.xfail(reason="TODO"),
            id="record_hexint"
        ),
        # TODO: quotes indicate transformer or grammar priority issue?
        pytest.param(
            """\
            record(ai, "value") {
                field(NAN, NaN)
                field(VAL, {
                    "a": 1e10
                })
            }
            """,
            None,
            marks=pytest.mark.xfail(reason="TODO"),
            id="record_misc_types"
        ),
        pytest.param(
            """\
            record(ai, "rec:X") {
                info(Q:group, {
                    "grp:name": {
                        "X": {
                            "+channel": "VAL"
                        }
                    }
                })
            }

            record(ai, "rec:Y") {
                info(Q:group, {
                    "grp:name": {
                        "Y": {
                            "+channel": "VAL"
                        }
                    }
                })
            }
            """,
            None,
            id="pva_group_1",
        ),
        pytest.param(
            """\
            record(ai, "time_tag") {
                info(Q:time:tag, "nsec:lsb:20")
            }

            record(ai, "display_form_hint") {
                info(Q:form, "Default")
            }

            record(longin, "tgt3") {
            }

            record(longin, "src3") {
                field(INP, {
                    "pva": {
                        "pv": "tgt3",
                        "field": "",
                        "local": false,
                        "Q": "4",
                        "pipeline": false,
                        "proc": "none",
                        "sevr": false,
                        "time": false,
                        "monorder": "0",
                        "retry": false,
                        "always": false,
                        "defer": false
                    }
                })
            }
            """,
            None,
            id="pva_group_2",
        ),
        pytest.param(
            """\
            record(longin, "src3") {
                field(INP, {
                    "pva": {
                        "Q": 4,
                        "proc": none,
                        "monorder": 0
                    }
                })
            }
            """,
            None,
            marks=pytest.mark.xfail(reason="TODO"),
            id="pva_group_3",
        ),
    ],
)
def test_roundtrip_format(input_text: str, expected: Optional[str]):
    input_text = textwrap.dedent(input_text)
    ctx = FormatContext(options=FormatOptions(indent=4))

    db = Database.from_string(input_text, version=4)
    result = ctx.render_object(db, "file")
    if expected is None:
        assert result.strip() == input_text.strip()
    else:
        expected = textwrap.dedent(expected)
        assert result.strip() == expected.strip()



import os

_true_values = {"1", "y", "yes", "true"}

# Maximum number of records to return in a single query for the frontend:
MAX_RECORDS = int(os.environ.get("WHATRECORD_MAX_RECORDS", 200))
# Path to the gdb binary to use for symbol introspection:
GDB_PATH = os.environ.get("WHATRECORD_GDB_PATH", "gdb")
# A path to cache parsed and serialized IOC information and other (e.g., gdb
# binary information):
CACHE_PATH = os.environ.get("WHATRECORD_CACHE_PATH", "")
# The default epics-base version to assume, where not specified otherwise.
DEFAULT_BASE_VERSION = os.environ.get("WHATRECORD_BASE_VERSION", "3.15")
# The server host to use (for the client)
WHATREC_SERVER = os.environ.get("WHATRECORD_SERVER", "http://localhost:8898/")
# List of plugins to enable.
PLUGINS = os.environ.get("WHATRECORD_PLUGINS", "happi twincat_pytmc netconfig epicsarch").split(" ")
# Period, in seconds, to scan for updated IOCs or files
SERVER_SCAN_PERIOD = int(os.environ.get("WHATRECORD_SERVER_SCAN_PERIOD", "600"))
# Period, in seconds, to reload autosave .sav files
AUTOSAVE_RELOAD_PERIOD = int(os.environ.get("WHATRECORD_AUTOSAVE_RELOAD_PERIOD", "60"))

# WHATRECORD_OUTPUT_INDENT (int) - indentation in spaces when rendering output.
INDENT = int(os.environ.get("WHATRECORD_OUTPUT_INDENT", "4"))

# WHATRECORD_MACRO_KEY_SKIP (str) - keys to skip, specified by regular
# expressions.  Should be valid Python syntax for a list of strings, as it
# will be handed to ``ast.literal_eval``.
MACRO_KEY_SKIP = os.environ.get("WHATRECORD_MACRO_KEY_SKIP", "")
# WHATRECORD_MACRO_INCLUDE_ENV (bool) - serialize environment variables if set
# to 1.
MACRO_INCLUDE_ENV = os.environ.get("WHATRECORD_MACRO_INCLUDE_ENV", "true").lower() in _true_values
# WHATRECORD_MACRO_VALUE_MAX_LENGTH (int) - maximum length of macro values to
# serialize. 0 to disable maximum length check.
MACRO_VALUE_MAX_LENGTH = int(os.environ.get("WHATRECORD_MACRO_VALUE_MAX_LENGTH", 1024))

# A SLAC-specific setting (other facilities may ignore this):
EPICS_SITE_TOP = os.environ.get("EPICS_SITE_TOP", "/reg/g/pcds/epics")



"""
"whatrecord deps" is used to get dependency information from EPICS IOC or
module Makefiles.

Under the hood, this uses GNU make, which is an external dependency required
for correct functionality.
"""

import argparse
import logging
from typing import List, Optional

from ..common import AnyPath
from ..makefile import DependencyGroup, DependencyGroupGraph, Makefile
from ..util import write_to_file
from .graph import render_graph_to_file

logger = logging.getLogger(__name__)
DESCRIPTION = __doc__


def build_arg_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    parser.description = DESCRIPTION
    parser.formatter_class = argparse.RawTextHelpFormatter

    parser.add_argument(
        "path", type=str, help="Path to IOC or Makefile itself"
    )

    parser.add_argument(
        "--keep-os-env",
        action="store_true",
        help=(
            "Keep environment variables present outside of ``make`` in the "
            ".env dictionaries"
        )
    )

    parser.add_argument(
        "--no-recurse",
        action="store_true",
        help="Do not recurse into dependencies",
    )

    parser.add_argument(
        "-of", "--output-format",
        default="json",
        type=str,
        help="Output format to use (json, console, file)",
    )

    parser.add_argument(
        "--graph",
        action="store_true",
        help="Output a graph of dependencies",
    )

    parser.add_argument(
        "-d",
        "--define",
        dest="defines",
        type=str,
        nargs="*",
        help="Define a variable for GNU make in the form: VARIABLE=VALUE",
    )

    parser.add_argument(
        "-o", "--graph-output",
        type=str,
        required=False,
        help="Graph output filename.  Defaults to standard output.",
    )

    return parser


def main(
    path: AnyPath,
    no_recurse: bool = False,
    keep_os_env: bool = False,
    graph: bool = False,
    graph_output: Optional[str] = None,
    output_format: str = "json",
    defines: Optional[List[str]] = None,
):
    variables = dict(variable.split("=", 1) for variable in defines or [])
    makefile_path = Makefile.find_makefile(path)
    makefile = Makefile.from_file(
        makefile_path, keep_os_env=keep_os_env, variables=variables
    )
    info = DependencyGroup.from_makefile(
        makefile, recurse=not no_recurse, keep_os_env=keep_os_env
    )

    if graph:
        group_graph = DependencyGroupGraph(info)
        render_graph_to_file(group_graph.to_digraph(), filename=graph_output)
        # An alternative to 'whatrecord graph'; both should have the same
        # result in the end.

    if not graph or graph_output is None:
        write_to_file(info, format=output_format)



"""Requires pytest-aiohttp"""
import json
import logging
from typing import (Any, Awaitable, Callable, Dict, Generator, Optional, Type,
                    TypeVar, Union)

import aiohttp
import aiohttp.test_utils
import aiohttp.web
import apischema
import pytest
import pytest_asyncio
from aiohttp.test_utils import BaseTestServer, TestClient, TestServer
from aiohttp.web import Application

from .. import gateway
from ..common import RecordInstance, WhatRecord
from ..server.server import (IocGetDuplicatesResponse, IocGetMatchesResponse,
                             IocGetMatchingRecordsResponse, PVGetInfo,
                             PVGetMatchesResponse, ServerHandler, ServerState,
                             _new_server)
from .conftest import MODULE_PATH, STARTUP_SCRIPTS

logger = logging.getLogger(__name__)

from .test_server_state import ready_state, state  # noqa

AiohttpClient = Callable[[Union[Application, BaseTestServer]], Awaitable[TestClient]]


@pytest.fixture()
def handler(ready_state: ServerState) -> ServerHandler:   # noqa: F811
    return ServerHandler(ready_state)


@pytest.fixture()
def server(handler: ServerHandler) -> aiohttp.web.Application:
    app = _new_server(handler, port=58433, run=False)
    # Don't do the initialization steps; we already have initialized enough
    # for the purposes of testing
    app.on_startup.remove(handler.async_init)
    return app


@pytest_asyncio.fixture
async def aiohttp_client() -> Generator[AiohttpClient, None, None]:
    """
    Factory to create a TestClient instance.

    Borrowed and modified from pytest-aiohttp, as a recent version is
    unavailable on conda-forge for our testing purposes.

    aiohttp_client(app, **kwargs)
    aiohttp_client(server, **kwargs)
    aiohttp_client(raw_server, **kwargs)
    """
    clients = []

    async def go(
        __param: Union[Application, BaseTestServer],
        *,
        server_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> TestClient:
        if isinstance(__param, Application):
            server_kwargs = server_kwargs or {}
            server = TestServer(__param, **server_kwargs)
            client = TestClient(server, **kwargs)
        elif isinstance(__param, BaseTestServer):
            client = TestClient(__param, **kwargs)
        else:
            raise ValueError(f"Unknown argument type: {type(__param)}")

        await client.start_server()
        clients.append(client)
        return client

    yield go

    while clients:
        await clients.pop().close()


@pytest_asyncio.fixture()
async def client(server: aiohttp.web.Application, aiohttp_client):
    return await aiohttp_client(server)


pvname = "IOC:KFE:A:One"
nonexistent_pvname = "pleasedonotmatch"
T = TypeVar("T")


async def get_response(
    client: aiohttp.web.Application,
    url: str,
    params: Optional[Dict[str, str]] = None
) -> aiohttp.ClientResponse:
    logger.debug("Request: %s (params=%s)", url, params)
    resp = await client.get(url, params=params or {})
    assert resp.status == 200
    return resp


async def get_json(
    client: aiohttp.web.Application, url: str,
    params: Optional[Dict[str, str]] = None
) -> dict:
    resp = await get_response(client, url, params)
    data = json.loads(await resp.text())
    logger.debug("Received: %s", data)
    return data


async def get_and_deserialize(
    client: aiohttp.web.Application, url: str, cls: Type[T],
    params: Optional[Dict[str, str]] = None
) -> T:
    data = await get_json(client, url, params)
    return apischema.deserialize(cls, data)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url, cls, params, expected",
    [
        pytest.param(
            "/api/pv/matches",
            PVGetMatchesResponse,
            dict(pattern=f"{pvname}.*", regex="true"),
            PVGetMatchesResponse(
                patterns=[pvname + ".*"],
                regex=True,
                matches=[pvname],
            ),
            id="match-regex"
        ),
        pytest.param(
            "/api/pv/matches",
            PVGetMatchesResponse,
            dict(pattern=f"{pvname}*", regex="false"),
            PVGetMatchesResponse(
                patterns=[pvname + "*"],
                regex=False,
                matches=[pvname],
            ),
            id="match-glob"
        ),
        pytest.param(
            "/api/ioc/pvs",
            IocGetMatchingRecordsResponse,
            dict(ioc="*", pv=nonexistent_pvname, regex="False"),
            IocGetMatchingRecordsResponse(
                ioc_patterns=["*"],
                record_patterns=[nonexistent_pvname],
                regex=False,
                matches=[],
            ),
            id="ioc-no-matching-records-glob"
        ),
        pytest.param(
            "/api/ioc/pvs",
            IocGetMatchingRecordsResponse,
            dict(ioc=".*", pv=nonexistent_pvname, regex="true"),
            IocGetMatchingRecordsResponse(
                ioc_patterns=[".*"],
                record_patterns=[nonexistent_pvname],
                regex=True,
                matches=[],
            ),
            id="ioc-no-matching-records-regex"
        ),
    ],
)
async def test_request(
    client,
    server: aiohttp.web.Application,
    url: str,
    cls: Type[T],
    params: Dict[str, str],
    expected: T,
):
    response = await get_and_deserialize(
        client,
        url=url,
        cls=cls,
        params=params,
    )
    assert response == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url, params",
    [
        pytest.param(
            "/api/ioc/pvs",
            dict(ioc="*", pv="*", regex="false"),
            id="match-glob"
        ),
        pytest.param(
            "/api/ioc/pvs",
            dict(ioc=".*", pv=".*", regex="true"),
            id="match-regex"
        ),
        pytest.param(
            "/api/file/info",
            dict(file=str(STARTUP_SCRIPTS[0])),
            id="file-info"
        ),
        pytest.param(
            "/api/logs/get",
            dict(),
            id="logs"
        ),
        pytest.param(
            "/api/pv/info",
            dict(pv=pvname),
            id="pv-info"
        ),
        pytest.param(
            "/api/pv/relations",
            dict(pv=pvname),
            id="relations"
        ),
        # No plugin info
        # pytest.param(
        #     "/api/plugin/info",
        #     dict(),
        #     id="plugin-info"
        # ),
    ],
)
async def test_request_smoke(
    client, server: aiohttp.web.Application,
    url: str,
    params: Dict[str, str],
):
    response = await get_json(
        client,
        url=url,
        params=params,
    )
    assert len(response)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url, params",
    [
        pytest.param(
            "/api/pv/graph",
            dict(pv=pvname, format="dot"),
            id="pv-dot"
        ),
        pytest.param(
            "/api/pv/script-graph",
            dict(pv=pvname, format="dot"),
            id="script-dot"
        ),
    ],
)
async def test_graph_smoke(
    client,
    server: aiohttp.web.Application,
    url: str,
    params: Dict[str, str],
):
    response = await get_response(
        client,
        url=url,
        params=params,
    )
    data = await response.text()
    assert "digraph" in data
    assert len(data)


@pytest.mark.parametrize(
    "pvname, key, value",
    [
        pytest.param(
            pvname,
            "gateway",
            {
                "name": "IOC:KFE:A:One",
                "matches": [
                    {
                        "filename": str(MODULE_PATH / "kfe.pvlist"),
                        "rule": {
                            "context": [
                                [
                                    str(MODULE_PATH / "kfe.pvlist"),
                                    10,
                                ]
                            ],
                            "pattern": "[A-Z][A-Z][A-Z]:KFE:.*",
                            "command": "ALLOW",
                            "header": "",
                            "metadata": {},
                        },
                        "groups": [],
                    }
                ],
            },
            id="gateway"
        ),
    ],
)
@pytest.mark.asyncio
async def test_record_metadata(
    client, server: aiohttp.web.Application,
    pvname: str,
    key: str,
    value: Any
):
    response = await get_and_deserialize(
        client,
        url="/api/pv/info",
        params=dict(pv=pvname),
        cls=Dict[str, PVGetInfo],
    )
    pv_get_info = response[pvname]
    assert pv_get_info.pv_name == pvname
    assert pv_get_info.present
    assert len(pv_get_info.info) == 1

    # Wow, does this really need to be so nested?!
    whatrec: WhatRecord = pv_get_info.info[0]
    instance: RecordInstance = whatrec.record.instance
    assert instance.metadata[key] == value


@pytest.mark.parametrize(
    "regex",
    [True, False]
)
@pytest.mark.asyncio
async def test_get_duplicates(
    client, server: aiohttp.web.Application, regex: bool
):
    response = await get_and_deserialize(
        client,
        url="/api/pv/duplicates",
        params=dict(regex=str(regex)),
        cls=IocGetDuplicatesResponse,
    )
    assert response == IocGetDuplicatesResponse(
        patterns=[".*" if regex else "*"],
        regex=regex,
        duplicates={"IOC:KFE:C:One": ["ioc_c", "ioc_d"]},
    )


@pytest.mark.asyncio
async def test_get_matches(
    client, server: aiohttp.web.Application,
):
    response = await get_and_deserialize(
        client,
        url="/api/ioc/matches",
        params=dict(pattern="ioc_[cd]$", regex="true"),
        cls=IocGetMatchesResponse,
    )
    assert response.patterns == ["ioc_[cd]$"]
    assert response.regex
    assert len(response.matches) == 2
    assert response.matches[0].name == "ioc_c"
    assert response.matches[1].name == "ioc_d"


@pytest.mark.asyncio
async def test_gateway_info(
    client, server: aiohttp.web.Application,
):
    info = await get_and_deserialize(
        client,
        url="/api/gateway/info",
        params=dict(),
        cls=gateway.GatewayConfig,
    )

    filename, kfe_pvlist = list(info.pvlists.items())[0]
    assert filename.name == "kfe.pvlist"
    assert kfe_pvlist.evaluation_order == "ALLOW, DENY"


# TODO: any way of testing plugin/nested/{info,keys}?



"""Smoke tests to see if the provided IOCs load without crashing."""

import pathlib

import pytest

from ..bin.graph import main
from . import conftest


@conftest.startup_scripts
def test_graph_smoke_startup_script(startup_script):
    main(filenames=[startup_script], highlight=[".*"])


@pytest.mark.parametrize(
    "with_dbd", [
        pytest.param(True, id="with_dbd"),
        pytest.param(False, id="without_dbd"),
    ]
)
def test_graph_db(with_dbd):
    db_path = conftest.MODULE_PATH / "iocs" / "streamdevice" / "test.db"
    if with_dbd:
        dbd_path = str(conftest.MODULE_PATH / "iocs" / "softIoc.dbd")
    else:
        dbd_path = None
    main(filenames=[db_path], dbd=dbd_path, highlight=[".*"])


@pytest.mark.parametrize(
    "with_dbd", [
        pytest.param(True, id="with_dbd"),
        pytest.param(False, id="without_dbd"),
    ]
)
def test_graph_multiple_db(with_dbd):
    db_paths = [
        conftest.MODULE_PATH / "iocs" / "streamdevice" / "test.db",
        conftest.MODULE_PATH / "iocs" / "ioc_a" / "ioc_a.db",
    ]
    if with_dbd:
        dbd_path = str(conftest.MODULE_PATH / "iocs" / "softIoc.dbd")
    else:
        dbd_path = None
    main(filenames=db_paths, dbd=dbd_path, highlight=[".*"])


@pytest.mark.parametrize(
    "filename",
    [
        # sequencer files
        conftest.MODULE_PATH / "iocs" / "ioc_sequencer" / "sncExample.st",
        # dependency graphs
        conftest.MODULE_PATH / "deps" / "module_a" / "Makefile",
    ]
)
def test_graph_other_formats_smoke(filename: pathlib.Path):
    main(filenames=[filename], highlight=[".*"])



"""dbLoadTemplate and msi -S substitution grammar helpers."""

from __future__ import annotations

import collections
import logging
import pathlib
import re
import shlex
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import apischema
import lark

from . import transformer
from .common import AnyPath, FullLoadContext
from .macro import MacroContext
from .transformer import context_from_token

logger = logging.getLogger(__name__)


def _strip_double_quote(value: str) -> str:
    """Strip one leading/single trailing double-quote."""
    if value[0] == '"':
        value = value[1:]
    if value[-1] == '"':
        value = value[:-1]
    return value


RE_REMOVE_ESCAPE = re.compile(r"\\(.)")


def _fix_value(value: str) -> str:
    """Remove quotes, and fix up escaping."""
    value = _strip_double_quote(value)
    return RE_REMOVE_ESCAPE.sub(r"\1", value)


@dataclass
class Substitution:
    """
    Single database template file from a full template.

    Represents approximately one line of a .substitutions file. For example,
    in this substitution file,

    ```
        file template.txt {
            pattern {a, b, c}
            {A, B, C}
        }
    ```

    The resulting Substitution would be

    ``Substitution(macros={"a": "A", "b": "B", "c": "C"}, filename="template.txt")``.

    Global macro values will be aggregated into this dictionary.

    Inside of the template file - ``template.txt`` above:
    * "include" is a supported command for the template file.
    * "substitute" is optionally supported (set ``allow_substitute``)
    """

    context: FullLoadContext
    filename: Optional[str] = None
    macros: Dict[str, str] = field(default_factory=dict)
    use_environment: bool = False
    allow_substitute: bool = True
    _items: Optional[List[Any]] = field(
        default_factory=list,
        metadata=apischema.metadata.skip,
    )

    def expand_file(
        self,
        *,
        filename: Optional[str] = None,
        search_paths: Optional[List[AnyPath]] = None
    ) -> str:
        """
        Expand the given file, looking in ``search_paths`` for template files.

        Parameters
        ----------
        filename : str, optional
            Expand this file or fall back to the instance-defined filename.

        search_paths : list of str or pathlib.Path, optional
            List of paths to search for template files.
        """
        filename = filename or self.filename
        if filename is None:
            raise ValueError("This substitution does not have a file defined")

        filename = pathlib.Path(filename)
        search_paths = search_paths or [filename.resolve().parent]
        with open(filename, "rt") as fp:
            return self.expand(fp.read(), search_paths=search_paths)

    @property
    def macro_context(self) -> MacroContext:
        """The macro context to be used when expanding the template."""
        ctx = MacroContext(use_environment=self.use_environment)
        ctx.define(**self.macros)
        return ctx

    @staticmethod
    def handle_include(filename: str, search_paths: List[AnyPath]) -> str:
        """Expand include files from the given search path."""
        for path in search_paths:
            option = pathlib.Path(path) / filename
            if option.exists():
                with open(option, "rt") as fp:
                    return fp.read()

        friendly_paths = " or ".join(f'"{path}"' for path in search_paths)
        raise FileNotFoundError(f"{filename} not found in {friendly_paths}")

    def expand(self, source: str, search_paths: Optional[List[AnyPath]] = None):
        """
        Expand the provided substitution template, using the macro environment.

        Parameters
        ----------
        source : str
            The source substitution template.  May contain "include" or
            "substitute" lines.

        search_paths : list of str or pathlib.Path, optional
            List of paths to search for template files.
        """
        ctx = self.macro_context
        search_paths = search_paths or [pathlib.Path(".")]
        results = []
        source_stack = collections.deque(source.splitlines())
        while source_stack:
            line = source_stack.popleft()
            logger.debug("line %r", line)
            line = ctx.expand(line)
            command, *command_args = line.strip().split(" ", 1)
            if command == "include":  # case sensitive
                args = shlex.split(command_args[0])
                if len(args) != 1:
                    raise ValueError(
                        f"Include command takes one argument; got: {args} "
                        f"where line={line!r}"
                    )
                include_file = args[0]
                logger.debug("Including file from %s", include_file)
                include_source = self.handle_include(include_file, search_paths)
                source_stack.extendleft(reversed(include_source.splitlines()))
                logger.debug("stack %r", source_stack)
            elif command == "substitute" and self.allow_substitute:
                # Note that dbLoadTemplate does not support substitute, but msi
                # does.
                macro_string = command_args[0].strip()
                # Strip only single beginning and end quotes
                macro_string = _strip_double_quote(macro_string).strip()
                logger.debug("Substituting additional macros %s", macro_string)
                ctx.define_from_string(macro_string)
            else:
                results.append(line)

        return "\n".join(results)


@dataclass
class VariableDefinitions:
    """Variable definitions."""

    context: FullLoadContext
    definitions: Dict[str, str] = field(default_factory=dict)


@dataclass
class GlobalDefinitions(VariableDefinitions):
    """Global variable definitions."""


@dataclass
class PatternHeader:
    """Pattern header."""

    context: FullLoadContext
    patterns: List[str]


@dataclass
class PatternValues:
    """Pattern values."""

    context: FullLoadContext
    values: List[str]


@dataclass
class TemplateSubstitution:
    """Database substitutions, containing zero or more template files."""

    substitutions: List[Substitution] = field(default_factory=list)

    def expand_template(
        self,
        template: str,
        search_paths: Optional[List[AnyPath]] = None,
        delimiter: str = "\n",
    ) -> str:
        """
        Expands all substitutions for the given string.

        Parameters
        ----------
        template : str
            The template text.

        delimiter : str, optional
            Delimiter to join individual substitutions.

        search_paths : list of str or pathlib.Path, optional
            List of paths to search for template files.
        """
        return delimiter.join(
            sub.expand(template, search_paths=search_paths)
            for sub in self.substitutions
        )

    def expand_files(
        self,
        search_paths: Optional[List[AnyPath]] = None,
        delimiter: str = "\n"
    ) -> str:
        """
        Expands and combines all contained substitution files.

        Parameters
        ----------
        delimiter : str, optional
            Delimiter to join individual substitutions.

        search_paths : list of str or pathlib.Path, optional
            List of paths to search for template files.
        """
        return delimiter.join(
            sub.expand_file(search_paths=search_paths)
            for sub in self.substitutions
        )

    @classmethod
    def from_string(
        cls,
        contents,
        filename=None,
        msi_format=False,
        all_global_scope=False,
    ) -> TemplateSubstitution:
        """Load a template substitutions file given its string contents."""
        comments = []
        grammar_filename = "msi-sub.lark" if msi_format else "dbtemplate.lark"

        grammar = lark.Lark.open_from_package(
            "whatrecord",
            grammar_filename,
            search_paths=("grammar",),
            parser="earley",
            # TODO: This is unsupported in lark:
            # lexer_callbacks={"COMMENT": comments.append},
            maybe_placeholders=False,
            propagate_positions=True,
        )

        if msi_format:
            tr = _TemplateMsiTransformer(
                cls, filename, all_global_scope=all_global_scope
            )
        else:
            tr = _TemplateTransformer(cls, filename)

        subs = tr.transform(grammar.parse(contents))
        subs.comments = comments
        return subs

    @classmethod
    def from_file_obj(cls, fp, filename=None) -> TemplateSubstitution:
        """Load a template file given a file object."""
        return cls.from_string(
            fp.read(),
            filename=getattr(fp, "name", filename),
        )

    @classmethod
    def from_file(cls, fn) -> TemplateSubstitution:
        """
        Load a template file.

        Parameters
        ----------
        filename : pathlib.Path or str
            The filename.

        Returns
        -------
        template : TemplateSubstitution
            The template file.
        """
        with open(fn, "rt") as fp:
            return cls.from_string(fp.read(), filename=fn)


@lark.visitors.v_args(inline=True)
class _TemplateMsiTransformer(lark.visitors.Transformer):
    _allow_substitute = True

    def __init__(self, cls, fn, visit_tokens=False, all_global_scope=False):
        super().__init__(visit_tokens=visit_tokens)
        self.fn = str(fn)
        self._template = cls()
        self._stack = []
        self._globals = {}
        self.all_global_scope = all_global_scope

    @lark.visitors.v_args(tree=True)
    def msi_substitutions(self, *_):
        self._template.substitutions = self._squash_stack(self._stack)
        return self._template

    template_filename = transformer.stringify
    pattern_name = transformer.stringify
    pattern_value = transformer.pass_through
    pattern_names = transformer.tuple_args
    variable_substitutions = transformer.tuple_args
    variable_definition = transformer.tuple_args
    variable_definitions = transformer.dictify

    def global_definitions(self, global_token: lark.Token, variable_definitions=None):
        self._stack.append(
            GlobalDefinitions(
                context=context_from_token(self.fn, global_token),
                definitions=dict(
                    (str(key), str(value))
                    for key, value in (variable_definitions or {}).items()
                ),
            )
        )

    def variable_subs(self, variable_definitions):
        token = list(variable_definitions)[0]
        definitions = VariableDefinitions(
            context=context_from_token(self.fn, token),
            definitions=dict(
                (str(key), _fix_value(value))
                for key, value in (variable_definitions or {}).items()
            ),
        )
        self._stack.append(definitions)
        return definitions

    def pattern_header(self, pattern_token: lark.Token, *values):
        header = PatternHeader(
            context=context_from_token(self.fn, pattern_token),
            patterns=list(_fix_value(value) for value in values),
        )
        self._stack.append(header)
        return header

    def pattern_values(self, *values):
        pattern_values = PatternValues(
            context=context_from_token(self.fn, values[0]),
            values=list(_fix_value(value) for value in values),
        )
        self._stack.append(pattern_values)
        return pattern_values

    @lark.visitors.v_args(tree=True)
    def empty(self, tree):
        empty = PatternValues(
            context=transformer.context_from_tree(self.fn, tree),
            values=[],
        )
        self._stack.append(empty)
        return empty

    def dbfile(self, file_token: lark.Token, filename: str, *fields):
        stack = self._stack
        file = Substitution(
            context=context_from_token(self.fn, file_token),
            filename=str(pathlib.Path(self.fn).parent / _strip_double_quote(filename)),
            allow_substitute=self._allow_substitute,
            _items=stack,
        )
        self._stack = [file]
        return file

    def _squash_stack(self, stack):
        patterns = []
        results = []
        for item in stack:
            if isinstance(item, Substitution):
                item_stack = item._items
                item._items = None
                subs = self._squash_stack(item_stack)
                for sub in subs:
                    sub.filename = item.filename
                results.extend(subs)
            elif isinstance(item, PatternHeader):
                patterns = item.patterns
            elif isinstance(item, PatternValues):
                values = dict(self._globals)
                pattern_dict = dict(zip(patterns, item.values))
                values.update(pattern_dict)
                results.append(
                    Substitution(
                        context=item.context,
                        filename=None,
                        macros=values,
                        allow_substitute=self._allow_substitute,
                        # fields=patterns,
                        # instances=values,
                    )
                )
            elif isinstance(item, GlobalDefinitions):
                self._globals.update(item.definitions)
            elif isinstance(item, VariableDefinitions):
                values = dict(self._globals)
                values.update(item.definitions)
                if self.all_global_scope:
                    self._globals.update(item.definitions)
                results.append(
                    Substitution(
                        context=item.context,
                        filename=None,
                        macros=values,
                        allow_substitute=self._allow_substitute,
                        # fields=patterns,
                        # instances=values,
                    )
                )

        return results


@lark.visitors.v_args(inline=True)
class _TemplateTransformer(_TemplateMsiTransformer):
    _allow_substitute = False

    @lark.visitors.v_args(tree=True)
    def substitution_file(self, *_):
        self._template.substitutions = self._squash_stack(self._stack)
        return self._template

    pattern_substitutions = transformer.tuple_args
    substitutions = transformer.tuple_args
